/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.jdbc;

import java.lang.*;
import java.sql.*;
import java.util.*;
import com.adobe.gl.runtime6.utils.GLException;

public class GLDataSource
	implements javax.sql.DataSource
{
	public GLDataSource(Hashtable a)
		throws ClassNotFoundException, InstantiationException,
		SQLException, IllegalAccessException
	{
		/* Set up member variables. */
		name = (String) a.get(GLDataSourceFactory.DSNAME);
		driver = (String) a.get("driver");
		dburl = (String) a.get("dburl");
		user = (String) a.get("user");
		password = (String) a.get("password");
		nstart = (String) a.get("pool.nstart");
		nmax = (String) a.get("pool.nmax");
		this.syn = makeSyndrome();
		log(10, "GLDataSource new " + syn + " {"/*}*/);

		/* Make sure driver is loaded into the JVM. */
		Class.forName(driver);

		/* Start a connection pool if requested. */
		int znstart = -1, znmax = -1;
		if ((znstart = getint(nstart, -1)) >= 0
		 && (znmax = getint(nmax, znstart)) >= znstart)
		{
			this.pool = new GLConnectionPool(this, znstart, znmax);
		}
		log(10, /*{*/"GLDataSource new }");
	}

	/**
	 ** Implement the javax.sql.DataSource interface.
	 **/

	public Connection
	getConnection()
		throws SQLException
	{
		log(10, "GLDataSource.get {"/*}*/);
		GLConnectionPool pool = this.pool;
		/*
		 * Race: if shutdown() gets in between our fetch of
		 * this.pool and pool.getConnection(), we will be
		 * talking to an orphaned pool.  But no big deal: the
		 * pool will notice it's an orphan and hand back a
		 * raw connection.
		 */
		Connection con = pool==null ? getcon()
			: new GLPooledConnection(pool.getConnection(), this);
		log(10, /*{*/"GLDataSource.get }");
		return con;
	}

	public Connection
	getConnection(String username, String password)
		throws SQLException
	{
		throw new SQLException("getConnection(u,p) not supported");
	}

	public int
	getLoginTimeout()
		throws SQLException
	{
		throw new SQLException("getLoginTimeout() not supported");
	}

	public java.io.PrintWriter
	getLogWriter()
		throws SQLException
	{
		throw new SQLException("getLogWriter() not supported");
	}

	public void
	setLogWriter(java.io.PrintWriter logger)
		throws SQLException
	{
		throw new SQLException("getLogWriter() not supported");
	}

	public void
	setLoginTimeout(int seconds)
		throws SQLException
	{
		throw new SQLException("setLoginTimeout(n) not supported");
	}

	/**
	 ** Other methods.
	 **/

	/**
	 * The characteristic function of this source.
	 */
	public String
	getSyndrome() {
		return syn;
	}

	/**
	 * Constructs the characteristic function of a source described
	 * by A.
	 */
	public static String
	makeSyndrome(Hashtable a) {
		String name, driver, dburl, user, password, nstart, nmax;

		name = (String) a.get(GLDataSourceFactory.DSNAME);
		driver = (String) a.get("driver");
		dburl = (String) a.get("dburl");
		user = (String) a.get("user");
		password = (String) a.get("password");
		nstart = (String) a.get("pool.nstart");
		nmax = (String) a.get("pool.nmax");

		return makeSyndrome(name, driver, dburl, user, password,
			nstart, nmax);
	}

	/**
	 * Constructs the characteristic function of this source.
	 * Caller can decide to shut down the data source based on this
	 * info.
	 */
	private String
	makeSyndrome() {
		return makeSyndrome(name, driver, dburl, user, password,
			nstart, nmax);
	}

	private static String
	makeSyndrome(String name, String driver, String dburl, String user,
		String password, String nstart, String nmax)
	{
		return new StringBuffer()
			.append("\n").append(GLDataSourceFactory.DSNAME)
				.append(" ").append(name)
			.append("\ndriver ").append(driver)
			.append("\ndburl ").append(dburl)
			.append("\nuser ").append(user)
			.append("\npassword ").append(password)
			.append("\nnstart ").append(nstart)
			.append("\nnmax ").append(nmax)
			.toString();
	}

	/**
	 * Return the connection to its pool.
	 */
	public void
	returnConnection(Connection con) {
		log(10, "GLDataSource.return {"/*}*/);

		GLConnectionPool pool = this.pool; /* synchronized access */
		if (con != null && pool != null) {
			pool.returnConnection(con);
			con = null;
		}

		/* Don't talk to DB inside sync block */
		if (con != null) {
			try {
				con.close();
			} catch (SQLException e) {}
		}

		log(10, /*{*/"GLDataSource.return }");
	}

	public Connection
	getcon()
		throws SQLException
	{
		return DriverManager.getConnection(dburl, user, password);
	}

	/**
	 * Stops connection pooling.
	 * {get,return}Connection will still work, but they'll just
	 * talk to the database each time.
	 */
	public void
	shutdown() {
		GLConnectionPool pool;
		final int mylev = 8;

		log(mylev, "GLDataSource.shutdown {"/*}*/);
		synchronized (this) {
			pool = this.pool;
			this.pool = null;
		}
		/* Don't talk to DB inside sync block */
		if (pool != null) {
			if (logp(mylev)) {
				StringBuffer ts = new StringBuffer();
				ts.append(syn).append("\n");
				pool.getStatistics(ts);
				log(mylev, ts.toString());
			}
			pool.closePool();
		}
		log(mylev, "GLDataSource.shutdown }");
	}

	/*
	 * -- Private --
	 */

	private int
	getint(String s, int val) {
		if (s == null)	/* no blame */
			return val;
		try {
			val = Integer.parseInt(s.trim());
		}
		catch (Throwable e) {
			log(1, "Couldn't decode ["+s+"] as an integer value.");
		}
		return val;
	}

	private void
	log(int thresh, String msg) {
		if (logLevel >= thresh)
			System.out.println(msg);
	}

	private boolean
	logp(int thresh) {
		return logLevel >= thresh;
	}

	private GLConnectionPool pool;
	private final String syn;
	private static final int logLevel = 9;
	private final String name, driver, dburl,
		user, password, nstart, nmax;
}
