/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.jdbc;

import java.io.*;
import java.net.*;
import java.sql.*;
import java.text.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import javax.servlet.jsp.*;
import com.adobe.gl.runtime6.utils.*;

/**
 * Action on a JDBC database.
 */
public class GLJdbc {
	private GLJdbc() {}

	/**
	 * Obtain a java.sql.Connection from the named DATABASE.
	 * The database name maps to an actual JDBC URL, driver,
	 * username and password via the application init parameters
	 * stored in web.xml.  For a database named "Magazine", these
	 * would be "com.adobe.gl.driver.Magazine",
	 * "com.adobe.gl.dburl.Magazine", etc.
	 */
	public static Connection
	getConnection(GLContext gl, String name)
		throws GLException
	{
		Connection con = null;
		try {
			con = GLDataSourceFactory.getInstance(gl, name)
				.getConnection();
			registerConnection(gl, con);
		}
		catch (SQLException e) {
			throw new GLException("SQL: "+e.toString());
		}

		if (con==null)
			throw new GLException("No connection for "+name);
		return con;
	}

	public static File
	getDataSourceFile(GLContext gl, String dstype) {
		final String fname = dstype.endsWith(JDBC_DSUF) ? dstype
			: dstype+JDBC_DSUF;
		File dir = new File(
			gl.configFileName(GLUtils.GL_PREFIX+"datasources."));
		gl.runtimeDebugMessage("ds directory " + dir.getAbsolutePath()
			+ " file " + fname);
		return new File(dir, fname);
	}

	/**
	 * Retrieves the list of known datasources
	 */
	public static String[]
	getDataSources(GLContext gl, boolean all) {
		return gl.getDataSources(JDBC_DSUF, all);
	}

	/**
	 * Return the number of rows in TABLENAME, determined by querying
	 * the database directly (not from a value cached in a result set).
	 */
	public static int
	getRecordCountBySelectQuery(Connection con, String tablename)
		throws SQLException
	{
		int n = 0;
		/*
		 * JDBC documents count(*), not count(0)
		 */
		PreparedStatement pst =
			con.prepareStatement("SELECT COUNT(*) FROM "
				+ tablename);
		ResultSet rs = pst.executeQuery();
		if (rs.next())
			n = rs.getInt(1);
		return n;
	}

	public static String
	getTableName(String sql)
		throws GLException
	{
		/*
		 * "select * from foobar" -> "foobar"
		 */
		StringTokenizer tk = new StringTokenizer(sql, " ");
		while (tk.hasMoreTokens())
			if ("FROM".equalsIgnoreCase(tk.nextToken())
			 && tk.hasMoreTokens())
				return tk.nextToken();
		throw new GLException("Can't decode " + sql);
	}

	/**
	 * Utility: build a clause suitable for inserting into a
	 * PreparedStatement template, using the field names in FIELDS.
	 */
	public static String
	makeBlankClause(String[] fields, String conjunction) {
		StringBuffer ts = new StringBuffer();
		String sep = "";
		for (int i = 0; i < fields.length; ++i, sep = conjunction)
			ts.append(sep).append(fields[i]).append("=? ");
		return ts.toString();
	}

	/**
	 * Return an unfilled WHERE clause:
	 *	"A=? AND B=? ..."
	 * Some of the items may be NULL, in which case we use
	 * "IS NULL" instead of "=?"; we need to modify the KEYS
	 * array to remove the key so that it doesn't participate
	 * in substitutions (@see this#fillClause()).
	 */
	public static String
	makeWhereClause(String[] keys, Hashtable vals) {
		StringBuffer ts = new StringBuffer();
		String sep = "";
		boolean nulls = false;
		for (int i = 0; i < keys.length; ++i, sep = " AND ") {
			ts.append(sep).append(keys[i]);
			if (GLNull.val.equals(vals.get(keys[i]))) {
				ts.append(" IS NULL");
				keys[i] = null;	/* remove this key */
				nulls = true;
			}
			else
				ts.append("=?");
		}
		/*
		 * Move nulls to the end of the array.
		 */
		if (nulls) {
			int n = 0;
			for (int i = 0; i < keys.length; ++i)
				if (keys[i] != null)
					keys[n++] = keys[i];
			for (int i = n; i < keys.length; ++i)
				keys[i] = null;
		}
		return ts.toString();
	}

	/**
	 * The KEYS array may have been modified (by makeWhereClause())
	 * to contain 1 or more trailing nulls.  Here return the number
	 * of non-null keys.
	 */
	public static int
	numKeys(String[] keys) {
		for (int n = keys.length; n > 0; --n)
			if (keys[n-1] != null)
				return n;
		return 0;
	}

	/**
	 * Reconstitute a Java object from a string representation S,
	 * guided by the jdbc type JTYPE.
	 */
	public static Object
	toObject(String s, int jtype, GLContext gl)
		throws GLException
	{
		switch (jtype) {
		case java.sql.Types.INTEGER:
		case java.sql.Types.TINYINT:
		case java.sql.Types.SMALLINT:
			return new Integer(s);
		case java.sql.Types.BIGINT:
			return new Long(s);
		case java.sql.Types.NUMERIC:
		case java.sql.Types.DECIMAL:
			return new java.math.BigDecimal(s);
		case java.sql.Types.BIT:
			return new Boolean(s);
		case java.sql.Types.CHAR:
		case java.sql.Types.VARCHAR:
		case java.sql.Types.LONGVARCHAR:
			return s;
		case java.sql.Types.REAL:
			return new Float(s);
		case java.sql.Types.FLOAT:
		case java.sql.Types.DOUBLE:
			return new Double(s);
		case java.sql.Types.DATE:
		case java.sql.Types.TIME:
		case java.sql.Types.TIMESTAMP:
			return toTimestamp(s, gl);
		}
		throw new GLException("Unsupported type "+jtype);
	}

	public static final String JDBC_DSUF = ".jdbc.sbs";

	/*
	 * -- Protected/private
	 */

	private static void
	registerConnection(GLContext gl, final Connection con) {
		gl.registerCloser(
			new GLNeedsCloser() {
				public void close() {
					try {
						con.close();
					}
					catch (SQLException e) {
					}
				}
			});
	}

	private static java.sql.Timestamp
	toTimestamp(String s, GLContext gl)
		throws GLException
	{
		java.sql.Timestamp stamp = null;
		long t = 0;
		try {
			stamp = java.sql.Timestamp.valueOf(s);
		} catch (IllegalArgumentException e) {}
		if (stamp != null)
			return stamp;

		try {
			java.util.Date d = new java.util.Date(s);
			stamp = new java.sql.Timestamp(d.getTime());
		} catch (IllegalArgumentException e) {}
		if (stamp != null)
			return stamp;

		try {
			java.util.Date d =
				new SimpleDateFormat("yyyy-MM-dd").parse(s);
			stamp = new java.sql.Timestamp(d.getTime());
		}
		catch (ParseException e) {
			throw new GLException(s + ": " + e.toString());
		}

		return stamp;
	}
};
