/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

/**
 * Superclass for most (not all) ContentSourceWrapper implementations.
 */
public abstract class CSWBase implements ContentSourceWrapper {
	protected CSWBase() {
		this.ix = 1;
		this.blksize = 1;
		this.nrow = 0;
		this.ncol = 0;
	}

	/**
	 * Current row.
	 * Row numbers start at 1.
	 */
	public int AbsolutePosition() {
		return ix;
	}
	public int BlockSize() {
		return blksize;
	}
	public void Close() {
	}
	/**
	 * Off the end of our block.
	 */
	public boolean EOB() {
		return EOF() || ix > lastRec;
	}
	/**
	 * True iff we've run off the end of the results.
	 */
	public boolean EOF() {
		return ix > nrow;
	}
	/**
	 * Usually override this (but not required)
	 */
	public String Error(String key) {
		return "";
	}
	public int FirstRecordOnPage() {
		return firstRec;
	}
	/*
	public java.util.Hashtable GetGenerator(boolean allRecords)
		throws GLException;
	 */
	public String GetName() {
		return name;
	}
	/*
	public String Key() throws GLException;
	 */
	public int LastRecordOnPage() {
		return lastRec;
	}
	public void Move(int delta) {
		ix += delta;
	}
	public void MoveFirst() {
		ix = 1;
	}
	public void MoveNext() {
		++ix;
	}
	public void MoveTo(int pos) {
		ix = pos;
	}
	/*
	void MoveToKey(java.util.Hashtable key) throws GLException;
	String[] PrimaryKey();
	 */
	/*
	ContentSourceWrapper NestedCSW(String name) throws GLException;
	 */
	public int RecordCount() {
		return nrow;
	}
	public String toString() {
		return new StringBuffer()
			.append("<p><b>CSW ").append(name).append("</b>\r\n")
			.append("ix ").append(ix)
			.append("<br>\r\nblksize ").append(blksize)
			.append("<br>\r\nnrow ").append(nrow)
			.append("<br>\r\nncol ").append(ncol)
			.append("<br>\r\nfirstRec ").append(firstRec)
			.append("<br>\r\nlastRec ").append(lastRec)
			.append("<br>\r\ncswerr {<br>\r\n").append(cswerr)
			.append("<br>}<br>\r\n")
			.toString();
	}
	/**
	 * Value of named column.
	 * Column names are case-insensitive.
	 *
	 * The ContentSourceWrapper provides both Value(String) and Value(int).
	 * Here we implement a scheme that links them by assigning numbers
	 * to column names.  But if you don't like it, just override the
	 * two Value methods.
	 */
	public Object Value(String key)
		throws GLException
	{
		return Value(getColumnByName(key));
	}

	protected int ix, blksize, nrow, ncol;
	protected int firstRec = 1, lastRec = 1;

	protected Object
	errorValue(int errix, String key)
		throws GLException
	{
		Object o = GLNull.val;
		if (cswerr != null) {
			cswerr.MoveTo(errix);
			o = cswerr.Value(key);
		}
		return o;
	}

	/**
	 * Must be overridden.
	 */
	protected int
	getColumnByName(String key) {
		return 0;
	}

	/**
	 * If there is an error content source available (in the
	 * session), and it has our name, installs it as this.cswerr.
	 * this.Value and this.Error consult this.cswerr.
	 */
	protected void
	initErrors(GLContext gl) {
		CSWError cse = gl.getError();
		if (cse==null)
			return;
		if (GetName()==null || !GetName().equals(cse.GetName()))
			return;
		this.cswerr = cse;
	}

	/**
	 * Checks row IX and column J against the current extents.
	 * If NEWRECORD is false, IX must be strictly within the extents;
	 * if NEWRECORD is true, IX may name a row 1 past the end.
	 * Returns RV_NO if invalid, RV_YES if valid, RV_NEW if we're
	 * in the new record case.
	 */
	protected int
	locValid(int ix, int j, boolean newRecord) {
		if (ix < 1 || j < 1 || this.ncol < j)
			return RV_NO;
		if (newRecord && this.nrow+1==ix)
			return RV_NEW;
		if (!newRecord && this.nrow < ix)
			return RV_NO;
		return RV_YES;
	}

	protected void
	setName(String name) {
		this.name = name;
	}

	protected static final int RV_NEW = 2, RV_YES = 1, RV_NO = 0;
	protected ContentSourceWrapper cswerr;
	protected String name;
};
