/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

import java.util.*;

/**
 * CSWBlockNavigator is a content source whose data is another content
 * source.  It is useful for moving around a large result set, e.g.,
 * a collection of search results that extends across many pages.
 * Its policy is to show a range of pages (from FirstRecordOnPage()
 * to LastRecordOnPage()); Next and Prev leap to the next/previous
 * page ranges.  E.g., when BlockSize is 5, and AbsolutePosition is 3,
 * it shows pages 1-5; on Next, it will show pages 6-10 (and set
 * AbsolutePosition to 6).
 */
public class CSWBlockNavigatorView
	extends CSWBase
	implements ContentSourceWrapper
{
	protected
	CSWBlockNavigatorView(GLContext gl, ContentSourceWrapper csw) {
		super();
		this.ncol = colname.length-1;
		this.gl = gl;
		this.viewed_csw = csw;
		int bs = csw.BlockSize();
		if (bs < 1)
			bs = 1;
		this.nrow = (csw.RecordCount() + bs - 1) / bs;
	}

	/**
	 * A new CSWBlockNavigatorView built from the given csw.
	 */
	public static CSWBlockNavigatorView
	newView(GLContext gl, String records, ContentSourceWrapper csw,
		String cname)
		throws GLException
	{
		CSWBlockNavigatorView csv = new CSWBlockNavigatorView(gl, csw);
		csv.setName(cname);
		csv.initPosition(records);
		if (gl.runtimeDebug)
			gl.runtimeDebugMessage(csv.toString());
		return csv;
	}

	/*
	 * The ContentSourceWrapper interface.
	 */

	public Hashtable
	GetGenerator(boolean allRecords)
		throws GLException
	{
		throw new GLException("GetGenerator");
	}

	public String
	Key()
		throws GLException
	{
		throw new GLException("Key");
	}

	public void
	MoveToKey(Hashtable key)
		throws GLException
	{
		throw new GLException("MoveToKey");
	}

	public ContentSourceWrapper
	NestedCSW(String name)
		throws GLException
	{
		throw new GLException(
			"CSWBlockNavigatorView: I don't do NestedCSW");
	}

	public String[]
	PrimaryKey()
	{
		return new String[0];
	}

	/**
	 * Value of numbered column.  The numbers are actually column
	 * names:
	 * "Block_Size" is the viewed csw's block size.
	 * "First" is the first record from the viewed csw's result
	 * set displayed on the current page.
	 * "Last" is the last record from the viewed csw's result
	 * set displayed on the current page.
	 * "Total_Records" is the viewed csw's RecordCount().
	 * "Link_URL" produces a navigation link.
	 * "Link_URL_checker" produces an empty string if the navigation
	 * link from "Link_URL" brings us back to the same record in the
	 * viewed_csw.
	 */
	public Object
	Value(int j)
		throws GLException
	{
		int bs = viewed_csw.BlockSize(),
			n = viewed_csw.RecordCount();
		switch (j) {
		case T_BLOCKSIZE:
			return new Integer(bs);
		case T_FIRST:
			return new Integer((ix-1)*bs+1);
		case T_LAST:
			int last = ix*bs;
			return new Integer(last < n ? last : n);
		case T_LINK_URL: {
			String urs = gl.linkToARecord(this, ix, true);
			return ufix(urs);
		}
		case T_LINK_URL_CHECKER: {
			String urs = gl.linkToARecord(this, ix);
			return "".equals(ufix(urs)) ? "" : "OK";
		}
		case T_TOTAL_RECORDS:
			return new Integer(n);
		default:
			break;
		}
		return "";
	}

	/*
	 * Other public functions
	 */

	/**
	 * Number of columns.
	 * Columns are numbered 1..getColumnCount().
	 */
	public int
	getColumnCount() {
		return ncol;
	}

	/**
	 * Name of column J.
	 * Columns are numbered 1..getColumnCount().
	 */
	public String
	getColumnName(int j) {
		return colname[j];
	}

	/**
	 * The CSW we are viewing.
	 */
	public ContentSourceWrapper
	getViewedCSW() {
		return viewed_csw;
	}

	public String
	toString() {
		return new StringBuffer()
			.append("<p><b>CSW ").append(name).append("</b>\r\n")
			.append(super.toString())
			.append("viewed_csw {<br>\r\n")
			.append(viewed_csw.toString())
			.append("<br>\r\n}<br>\r\n")
			.toString();
	}

	/*
	 * -- private --
	 */

	protected int
	getColumnByName(String fieldName) {
		for (int j = ncol; j > 0; --j)
			if (colname[j].equalsIgnoreCase(fieldName))
				return j;
		return 0;
	}

	private void
	initPosition(String records)
		throws GLException
	{
		GLBlockParam pb = new GLBlockParam(records);
		if (!pb.valid) {
			throw new GLException(
				"CSWBlockNavigator.newView: unknown arg "
				+ records);
		}
		this.blksize = pb.blksize==0 ? this.nrow : pb.blksize;
		int bs = viewed_csw.BlockSize();
		if (bs < 1)
			bs = 1;
		int i = (viewed_csw.FirstRecordOnPage() + bs - 1) / bs;
		this.MoveTo(i);
		this.firstRec = this.blksize==0 ? 1
			: this.ix - (this.ix-1) % this.blksize;
		this.lastRec = this.firstRec + this.blksize - 1;
		if (this.lastRec > this.nrow)
			this.lastRec = this.nrow;
	}

	/**
	 * Retrieves the value of an href attribute U.
	 * Specifically, if U matches " href=['"](.*)['"]", returns the
	 * substring between (); otherwise returns U unchanged.
	 * Could be more robust (varying leading whitespace, href
	 * capitalization, no quotes around attribute value);
	 * but works with the result of calling gl.linkToARecord().
	 */
	private String
	ufix(String u) {
		final String hprefix = " href=";
		final int nhprefix = hprefix.length();
		gl.runtimeDebugMessage("ufix "+u);
		if (u.length() <= nhprefix || !u.startsWith(hprefix))
			return u;
		char c = u.charAt(nhprefix);
		int lx = u.indexOf(c, 1+nhprefix);
		if (lx < 0)
			lx = u.length();
		u = u.substring(1+nhprefix, lx);
		gl.runtimeDebugMessage("ufixed "+u);
		return u;
	}

	private GLContext gl;
	private ContentSourceWrapper viewed_csw;

	private static final int
		T_BLOCKSIZE = 1,
		T_FIRST = 2,
		T_LAST = 3,
		T_LINK_URL = 4,
		T_LINK_URL_CHECKER = 5,
		T_TOTAL_RECORDS = 6;
	private static final String[] colname = {
		"",
		"Block_Size",
		"First",
		"Last",
		"Link_URL",
		"Link_URL_checker",
		"Total_Records",
	};
};
