/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

import java.util.*;
import org.apache.regexp.*;

/**
 * Error content source.
 * When the user posts a form, the form's action page can choose to
 * populate a CSWError object and send it along (by installing it into
 * the session) after a redirect.
 * The expected scenario is that when the action page encounters an error
 * (invalid input, database update failure), it redirects to the original
 * form page with a CSWError set.  The form page redisplays itself from
 * a content source formed from the union of its original content source
 * and the CSWError.  So any input the user typed into the form before
 * posting will be displayed again (courtesy of the CSWError) and he
 * gets a chance to correct the invalid input.  Additionally, error
 * messages retrieved from the CSWError can be used to decorate the
 * form (e.g., with style attributes to mark which fields have errors.)
 */
public class CSWError extends CSWBase implements ContentSourceWrapper {
	public CSWError() {
		super();
		lastRec = 0;
		firstRec = 630507;	/* happy birthday jma */
	}

	/**
	 * A new CSWError built from the posted form data.
	 */
	public static CSWError
	newError(GLContext gl, String records) {
		CSWError cse = new CSWError();
		cse.loadFormData(gl);
		cse.loadDollarDataFromForm(gl);
		String m = (String) cse.dollar.get("$cswName");
		if (m != null)
			cse.setName(m);
		/* ignore RECORDS arg */
		cse.MoveTo(cse.FirstRecordOnPage());
		if (gl.runtimeDebug)
			gl.runtimeDebugMessage(cse.toString());
		return cse;
	}

	/*
	 * The ContentSourceWrapper interface.
	 */

	/**
	 * The error message associated with field FIELDNAME in the
	 * current row, or <code>""</code> if none exists.
	 */
	public String
	Error(String fieldName) {
		int j = getColumnByName(fieldName);
		String m = null;
		if (locValid(ix, j, false) == RV_YES)
			m = (String) errors.get(ekey(ix, j));
		return m==null ? "" : m;
	}

	public Hashtable
	GetGenerator(boolean allRecords)
		throws GLException
	{
		throw new GLException("GetGenerator");
	}

	public String
	Key()
		throws GLException
	{
		throw new GLException("Key");
	}

	public void
	MoveToKey(Hashtable key)
		throws GLException
	{
	Rowloop:
		for (MoveTo(FirstRecordOnPage()); !EOB(); Move(1)) {
			for (Enumeration e = key.keys(); e.hasMoreElements();) {
				String k = (String) e.nextElement();
				if (!key.get(k).equals(Value(k)))
					continue Rowloop;
			}
			return;
		}
	}

	public ContentSourceWrapper
	NestedCSW(String name)
		throws GLException
	{
		throw new GLException("CSWError: I don't do NestedCSW");
	}

	public String[]
	PrimaryKey()
	{
		return new String[0];
	}

	/**
	 * Value of named column.
	 * Column names are case-insensitive.
	 */
	public Object
	Value(String key)
		throws GLException
	{
		return Value(getColumnByName(key));
	}

	public Object
	Value(int j)
		throws GLException
	{
		Object o = null;
		if (locValid(ix, j, false) == RV_YES)
			o = data.get(ekey(ix, j));
		return GLNull.nonnull(o);
	}

	/*
	 * Other public functions
	 */

	/**
	 * Replaces a data value VAL associated
	 * with field FIELDNAME in the current row.
	 * <p>Ignores out-of-range rows and FIELDNAMEs.
	 */
	public void
	ChangeValue(String fieldName, Object val) {
		doSet(data, fieldName, val);
	}

	/**
	 * Returns a reference to the internal dollar-data hash.
	 */
	public Hashtable
	DollarData() {
		return dollar;
	}

	public int
	getColumnCount() {
		return ncol;
	}

	public String
	getColumnName(int j) {
		Integer jz = new Integer(j);
		for (Enumeration e = colnamehash.keys(); e.hasMoreElements();) {
			String name = (String) e.nextElement();
			if (jz.equals(colnamehash.get(name)))
				return name;
		}
		return null;
	}

	/**
	 * Installs (or replaces) an error message ERRSTR associated
	 * with field FIELDNAME in the current row.
	 * <p>Ignores out-of-range rows and FIELDNAMEs.
	 */
	public void
	SetError(String fieldName, String errstr) {
		doSet(errors, fieldName, errstr);
	}

	public String
	toString() {
		return new StringBuffer()
			.append("<p><b>CSW ").append(name).append("</b>\r\n")
			.append(super.toString())
			.append("data {<br>\r\n")
			.append(data.toString())
			.append("<br>\r\n}<br>\r\n")
			.append("dollar {<br>\r\n")
			.append(dollar.toString())
			.append("<br>\r\n}<br>\r\n")
			.append("errors {<br>\r\n")
			.append(errors.toString())
			.append("<br>\r\n}<br>\r\n")
			.toString();
	}

	/*
	 * -- private --
	 */

	private Hashtable colnamehash = new Hashtable();

	/**
	 * Installs (or replaces) an object VAL into H associated with
	 * the current row and FIELDNAME field.
	 * Ignores out-of-range rows and FIELDNAMEs.
	 */
	private void
	doSet(Hashtable h, String fieldName, Object val) {
		int j = getColumnByName(fieldName);
		if (locValid(ix, j, false) != RV_YES)
			return;
		final String key = ekey(ix, j);
		if (val == null)
			h.remove(key);
		else
			h.put(key, val);
	}

	/**
	 * Standard key for current row and column.
	 */
	private static String
	ekey(int ix, int j) {
		return "row "+ix+" col "+j;
	}

	/**
	 * Ensures that there is a row IX and field FIELDNAME.
	 */
	private void
	expandExtents(int ix, String fieldName) {
		if (ix < 1)
			return;
		if (this.firstRec > ix)
			this.firstRec = ix;
		if (this.lastRec < ix)
			this.lastRec = ix;
		if (this.nrow < ix)
			this.nrow = ix;
		this.blksize = this.lastRec - this.firstRec + 1;
		if (colnamehash.get(fieldName) == null)
			colnamehash.put(fieldName, new Integer(++this.ncol));
	}

	protected int
	getColumnByName(String fieldName) {
		Integer b = (Integer) colnamehash.get(fieldName);
		return b==null ? 0 : b.intValue();
	}

	protected void
	initErrors(GLContext gl) {
	}

	/**
	 * Finds form parameters whose names begin with '$' -- these
	 * are assumed to be special control flags for the action --
	 * and installs them into the DOLLAR hash available via
	 * this.dollarData().
	 */
	private void
	loadDollarDataFromForm(GLContext gl) {
		Hashtable h = gl.getRequestParameters();
		for (Enumeration e = h.keys(); e.hasMoreElements();) {
			String name = (String) e.nextElement();
			if (!name.startsWith("$"))
				continue;
			dollar.put(name, h.get(name));
		}
	}

	/**
	 * Looks for request parameters like FOO[IX]
	 * and installs each one as field FOO in row IX of
	 * our content source, later accessible via
	 * <code>MoveTo(IX); Value(FOO)</code>
	 */
	private void
	loadFormData(GLContext gl) {
		RE re = null;
		try {
			re = new RE("^~?(.+)\\[(\\d+)\\]$");
		}
		catch (RESyntaxException e) {
			gl.runtimeDebugMessage("Couldn't parse RE: "+e);
		}

		Hashtable form = gl.getRequestParameters();
		for (Enumeration e = form.keys(); e.hasMoreElements();) {
			String key = (String) e.nextElement();
			if (!re.match(key))
				continue;
			String column = re.getParen(1);
			String six = re.getParen(2);
			int ix = 0;
			try {
				ix = Integer.parseInt(six);
			} catch (Exception ez) { /* "cannot happen" */ }
			expandExtents(ix, column);
			MoveTo(ix);
			ChangeValue(column, gl.getFormValue(form, key));
		}
		if (this.firstRec == 630507) {
			this.firstRec = 1;
			this.lastRec = 0;
			this.nrow = 0;
			this.blksize = 0;
		}
	}

	private Hashtable data = new Hashtable();
	private Hashtable dollar = new Hashtable();
	private Hashtable errors = new Hashtable();
};
