/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

import java.util.*;
import org.apache.regexp.*;

/**
 * Conditions.
 * The GoLive code (at user direction) writes tests into the page source
 * which are evaluated at JSP execution time to determine whether the
 * page content controlled by the test should be shown or hidden.  The
 * test predicates are defined here; GoLive learned their names by looking
 * for &lt;condition>...&lt;/condition> elements in the UI XML files.
 */
public class GLCondition {
	public static boolean
	FirstRecord(GLContext gl, ContentSourceWrapper csw) {
		return csw.FirstRecordOnPage()==1;
	}
	public static boolean
	HasError(GLContext gl, Object o) {
		return !"".equals(o);
	}
	public static boolean
	IsEmpty(GLContext gl, Object o) {
		return "".equals(GLNull.nonnull(o).toString());
	}
	public static boolean
	IsTrue(GLContext gl, boolean o) {
		return o;
	}
	public static boolean
	IsTrue(GLContext gl, Object o) {
		return gl.truth(o);
	}
	public static boolean
	LastRecord(GLContext gl, ContentSourceWrapper csw) {
		return csw.LastRecordOnPage()==csw.RecordCount();
	}
	public static boolean
	MatchField(GLContext gl, Object o, String cond, Object value)
		throws GLException
	{
		return MatchField(gl, o, cond, GLNull.nonnull(value).toString());
	}
	public static boolean
	MatchField(GLContext gl, Object o, String cond, String value)
		throws GLException
	{
		if ("==".equals(cond))
			return icmp(gl, o, value) == 0;
		if ("!=".equals(cond))
			return icmp(gl, o, value) != 0;
		if ("<".equals(cond))
			return icmp(gl, o, value) < 0;
		if (">".equals(cond))
			return icmp(gl, o, value) > 0;
		if ("<=".equals(cond))
			return icmp(gl, o, value) <= 0;
		if (">=".equals(cond))
			return icmp(gl, o, value) >= 0;
		if ("matches regexp".equals(cond)) {
			try {
				return re_match(o, value);
			}
			catch (RESyntaxException e) {
				gl.runtimeErrorMessage(
					"Syntax error in " + value);
			}
			return false;
		}
		throw new GLException("unknown MatchField op: "+cond);
	}
	public static boolean
	NoRecords(GLContext gl, ContentSourceWrapper csw) {
		return 0==csw.RecordCount();
	}
	public static boolean
	RecordCountGreaterThan(GLContext gl, ContentSourceWrapper csw, int n) {
		return n < csw.RecordCount();
	}
	public static boolean
	ReturnFalse(GLContext gl) {
		return false;
	}
	public static boolean
	ReturnTrue(GLContext gl) {
		return true;
	}

	/**
	 * Returns -1,0,1 as O is considered less than, equal to,
	 * or greater than VALUE.
	 */
	private static int
	icmp(GLContext gl, Object o, String value) {
		int t = 0;
		if (gl.runtimeDebug)
			gl.runtimeDebugMessage("<br>icmp o " + o + " value " + value);
		if (GLNull.val.equals(o))
			o = "";

		if (o instanceof Number) {
			try {
				return icmpNumeric(gl, (Number) o, value);
			}
			catch (NumberFormatException e) {
			}
		}

		if (o instanceof Boolean) {
			int i = ((Boolean) o).booleanValue() ? 1 : 0;
			int j = "true".equalsIgnoreCase(value)
				|| "1".equalsIgnoreCase(value) ? 1 : 0;
			if (gl.runtimeDebug)
				gl.runtimeDebugMessage("Boolean compare => " + (i-j));
			return i - j;
		}
		if (o instanceof String) {
			t = ((String) o).compareTo(value);
			if (gl.runtimeDebug)
				gl.runtimeDebugMessage("String compare => " + t);
			return t;
		}
		t = o.toString().compareTo(value);	/* hmmm... */
		if (gl.runtimeDebug)
			gl.runtimeDebugMessage("default => " + t);
		return t;
	}

	/**
	 * Attempts to compare O numerically to VALUE by converting
	 * VALUE to one of the Number concrete subtypes.
	 * Returns -1,0,1 as O is less than, equal to, or greater than
	 * VALUE; or throws NumberFormatException if it cannot parse
	 * a number out of VALUE.
	 */
	private static int
	icmpNumeric(GLContext gl, Number o, String value)
		throws NumberFormatException
	{
		int t = 0;

		if (o instanceof java.math.BigDecimal) {
			java.math.BigDecimal d1 = (java.math.BigDecimal) o;
			java.math.BigDecimal d2 = new java.math.BigDecimal(value);
			t = d1.compareTo(d2);
			if (gl.runtimeDebug) {
				gl.runtimeDebugMessage("BigDecimal compare => " + t
					+ " d1: " + d1 + " d2: " + d2);
			}
			return t;
		}
		/*
		 * A database "float" value is better (i.e., has
		 * less surprising roundoff error) as a Float than
		 * as a Double.
		 */
		if (o instanceof Float) {
			float d1 = o.floatValue();
			float d2 = Float.parseFloat(value);
			t = d1 < d2 ? -1 : d1 > d2 ? 1 : 0;
			if (gl.runtimeDebug) {
				gl.runtimeDebugMessage("Float compare => " + t
					+ " d1: " + d1 + " d2: " + d2);
			}
			return t;
		}
		/*
		 * IEEE doubles represent integers exactly to something
		 * like +/-2**53.  So we get all the integral types except
		 * Long, for which we won't have exact values in the extreme
		 * range.
		 */
		double d1 = o.doubleValue();
		double d2 = Double.parseDouble(value);
		t = d1 < d2 ? -1 : d1 > d2 ? 1 : 0;
		if (gl.runtimeDebug) {
			gl.runtimeDebugMessage("Double compare => " + t
				+ " d1: " + d1 + " d2: " + d2);
		}
		return t;
	}

	/**
	 * Returns TRUE/FALSE as O matches/does not match VALUE,
	 * or throws RESyntaxException if VALUE cannot be
	 * compiled into a regular expression.
	 */
	private static boolean
	re_match(Object o, String value)
		throws RESyntaxException
	{
		return new RE(value).match(GLNull.nonnull(o).toString());
	}
};
