/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

import javax.servlet.*;
import javax.servlet.http.*;
import javax.servlet.jsp.*;
import java.beans.*;
import java.io.*;
import java.net.*;
import java.sql.*;
import java.util.*;

/**
 * Unlike PHP and ASP/VBscript, Java doesn't offer any "global" variables
 * to reference the executing page's request, response, session, and
 * application data.  So GoLive generates one of these gizmos
 * at the top of every page, and we pass it around as needed.
 */
public class GLContext {
	/**
	 * The JSP PageContext object has methods to get at the
	 * request, response, &c. objects.
	 */
	public GLContext(PageContext pc) {
		this.pc = pc;
		this.starttime = System.currentTimeMillis();

		this.runtimeDebug = getDebug(getApplication());
		this.doRecode = getDoRecode(getApplication());
		this.id = nextID();
		this.idstr = new StringBuffer().append(this.firststart)
			.append('#')
			.append(GLNumberFormat.Number(this,
				new Integer(this.id), 0, 0, 4, 0))
			.toString();

		/*
		 * Make ourselves available to custom tags.
		 */
		pc.setAttribute(GL_NAME, this, pc.PAGE_SCOPE);

		HttpServletRequest req = this.getRequest();
		HttpSession ssn = req.getSession(false);
		friendly = ssn != null
			&& this.truth(ssn.getAttribute(friendlyName));
		runtimeDebugMessage("<p>START "+req.getRequestURI());
		if (req.getAttribute(idstrName) == null)
			req.setAttribute(idstrName, idstr);
	}

	/**
	 * "Show Details of Empty Record"
	 */
	public String
	AddNewRecordKey(ContentSourceWrapper csw) {
		String rkname = GLUtils.rk(csw.GetName()),
			rkval = "$newRecord";
		return new StringBuffer(URLEncoder.encode(rkname))
			.append('=')
			.append(URLEncoder.encode(rkval))
			.toString();
	}

	/**
	 * "Show Details of Current Record"
	 */
	public String
	AddRecordKey(ContentSourceWrapper csw) {
		String rkname = GLUtils.rk(csw.GetName()),
			rkval = null;
		try {
			rkval = csw.Key();
		}
		catch (Exception e) {
			runtimeDebugMessage("Caught "+e.toString());
			rkval = URLEncoder.encode("$newRecord");
		}
		return new StringBuffer(URLEncoder.encode(rkname))
			.append('=')
			.append(rkval)	/* NB already encoded */
			.toString();
	}

	/**
	 * Returns either "" or an HTML attribute like ANAME='value'
	 * where value is computed from ASPEC and POS.
	 * POS is a 1-based selector into the array described by ASPEC.
	 * ASPEC is a (comma-separated) list of number-value pairs that
	 * specifies how the alternation works.  An ASPEC like
	 * "2:white,2:grey" describes 2 white items, followed by 2 grey
	 * items, (then repeat): if pos is 1 or 2 (or 5 or 6, or 9 or 10,
	 * ...), value is white; if pos is 3 or 4 (or 7 or 8, ...), then
	 * value is grey.
	 * <p>
	 * The special value "_no_attr" means return an empty string,
	 * not ANAME='_no_attr'.
	 */
	public String
	alternate(int pos, String aname, String aspec) {
		/* One-deep cache. */
		Vector v = aspec.equals(this.aspec) ? this.avec : null;
		if (v==null) {	/* Miss: generate avec. */
			String[] specs = GLUtils.split(",", aspec);
			v = new Vector();
			for (int i = 0; i < specs.length; ++i) {
				int sepx = specs[i].indexOf(":");
				int z = Integer.parseInt(
					specs[i].substring(0, sepx));
				String w = specs[i].substring(1+sepx);
				while (--z >= 0)
					v.add(w);
			}
			this.aspec = aspec;
			this.avec = v;
		}

		final int N = v.size();
		if (N == 0)
			return "";
		String val = (String) v.get(((pos-1) % N + N) % N);
		if ("_no_attr".equals(val))
			return "";
		return GLUtils.cat(GLUtils.htmlEncode(aname), "='",
			GLUtils.htmlEncode(val), "'");
	}

	/**
	 * "checked"/"" as COND is true/false.
	 */
	public static String
	checked(boolean cond) {
		return cond ? "checked" : "";
	}

	/**
	 * Does what we need to do to leave this page -- called
	 * from <code>finally</code> blocks, so it catches both normal
	 * and abrupt exits.
	 * <p>
	 * Most important, shuts down things like database connections.
	 */
	public void
	closeAll() {
		for (int i = needsClosing.size(); --i >= 0;)
			((GLNeedsCloser) needsClosing.get(i)).close();
	}

	/**
	 * Returns the file (or directory) name associated with KEY.
	 * KEY looks like "com.adobe.gl.friends" or "com.adobe.gl.datasources.":
	 * segments, separated by ".", perhaps with a "." at the end.
	 * Without the trailing ".", the caller is requesting a filename:
	 * if we find it, we replace instances of "?" in the returned filename
	 * with the last segment of KEY.
	 * With the trailing ".", the caller is requesting a directory name,
	 * planning to do its own filename filtering.
	 * <p>
	 * A note on finding: passed a key like "com.foo.bar", we
	 * will successively search the internal table for "com.foo.bar"
	 * "com.foo.", and "com." until we find a match or exhaust the
	 * sequence.
	 */
	public String
	configFileName(String key)
	{
		runtimeDebugMessage("configFileName("+key+")");
		/*
		 * Discover VAL, the value of KEY, and BASENAME, the
		 * rightmost '.'-separated segment of VAL.
		 * - VAL may be stored directly in cfmap.
		 * - VAL may be null if we didn't find anything.
		 * - A prefix of KEY may be present in the map.
		 *   We extract the value and compute VAL by appending
		 *   the remaining segments of KEY.
		 */
		final String basename;
		String val = (String) cfmap.get(key);
		if (val != null) {
			int ix = val.lastIndexOf(FSP);
			int dotix = key.lastIndexOf(".");
			basename = dotix+1==key.length() ? ""
				: ix < 0 ? val
				: val.substring(ix+1);
		}
		else {
			String[] a = GLUtils.split(".", key);
			basename = a.length == 0 ? "" : a[a.length-1];
			for (int n = a.length; --n > 0;) {
				String k = GLUtils.join(".", a, 0, n) + ".";
				if ((val = (String) cfmap.get(k)) != null) {
					val += GLUtils.join(FSP, a,
						n+1, a.length);
					break;
				}
			}
		}

		if (val != null) {
			/*
			 * If BASENAME isn't empty, substitute it for '?';
			 * otherwise, trim off the last FSP-separated
			 * segment of VAL (i.e., get its parent directory).
			 */
			if (basename.length() > 0) {
				String[] a = GLUtils.split("?", val);
				val = GLUtils.join(basename, a);
			}
			else {
				String[] a = GLUtils.split(FSP, val);
				a[a.length-1] = "";
				val = GLUtils.join(FSP, a);
			}
		}

		/*
		 * This mechanism only works with files inside the
		 * web application.
		 */
		if (val != null) {
			String rp = getApplication().getRealPath("/");
			if (!rp.endsWith(FSP))
				val = rp + FSP + val;
			else
				val = rp + val;
		}
		return val;
	}

	/**
	 * Tries to convert URL (an absolute, rooted, or page-relative URL)
	 * into a context-relative URL.
	 * If successful, the returned URL has a leading '/', and
	 * the caller can use the internal getResourceAsStream() to
	 * fetch bytes.
	 * Otherwise, the returned URL can be accessed via an HTTP request.
	 */
	public String
	contextPathFrom(String url) {
		HttpServletRequest req = getRequest();
		String full = HttpUtils.getRequestURL(req).toString();
		int k;

		/*
		 * Absolute URL.  If it shares our scheme://host:port,
		 * turn it into a rooted URL and proceed; otherwise, done.
		 */
		if ((k = url.indexOf("://")) > 0) {
			int l = url.indexOf("/", k+3);
			if (l < 0)
				l = url.length();
			String ohost = url.substring(0, l+1);
			if (!full.startsWith(ohost))
				return url;
			url = l==url.length() ? "/" : url.substring(l);
		}
		/*
		 * Rooted URL.  If it's not in the current application
		 * context, prepend scheme://host:port.
		 */
		if (url.indexOf('/') == 0) {
			String me = req.getContextPath();
			if (url.startsWith(me))
				url = url.substring(me.length());
			else {
				k = full.indexOf("://");
				int l = full.indexOf("/", k+3);
				if (l < 0) l = full.length();
				String fhost = full.substring(0, l);
				return fhost + url;
			}
		}
		/* Page-relative URL. */
		if (url.indexOf('/') != 0) {
			String dir = req.getServletPath();
			int diri = dir.lastIndexOf('/');
			if (diri >= 0 && diri < dir.length()-1)
				dir = dir.substring(0, diri);
			url = dir + "/" + url;
		}
		return url;
	}

	/**
	 * Allows GoLive to address cookies by name.
	 */
	public Object
	Cookie(String s) {
		Cookie[] cookies = getRequest().getCookies();
		for (int i = cookies.length; --i >= 0;)
			if (s.equalsIgnoreCase(cookies[i].getName()))
				return cookies[i];
		return null;
	}

	/**
	 * Compares O1, O2 as database values.
	 */
	public static boolean
	eqdbval(Object o1, Object o2) {
		return o1==o2 || o1!=null && o1.equals(o2);
	}

	/**
	 * Compares O as a string to S.
	 */
	public static boolean
	eqstr(Object o, String s) {
		return GLNull.nonnull(o).toString().equals(s);
	}

	/**
	 * Protects quote characters (e.g., inside HTML attribute values):
	 *	&lt;a alt="Rocco &amp;#34;Slugger&amp;#34; O&amp;#39;Reilly"
	 *		href="...">
	 */
	public static String
	fixHTMLquotes(Object os) {
		String s = GLNull.nonnull(os).toString();
		final String q1 = "&#39;", q2 = "&#34;";
		StringBuffer w = new StringBuffer();
		final int NS = s.length();
		for (int i = 0; i < NS; ++i) {
			final char c = s.charAt(i);
			if (c=='\'')
				w.append(q1);
			else if (c=='"')
				w.append(q2);
			else
				w.append(c);
		}
		return w.toString();
	}

	/**
	 * This ServletContext corresponds to the "application"
	 * local variable built into a JSP page.
	 */
	public ServletContext
	getApplication() {
		return pc.getServletContext();
	}

	/**
	 * Retrieves the list of known datasources of type SUF.
	 */
	public String[]
	getDataSources(String suf, boolean all) {
		File dir = new File(
			configFileName(GLUtils.GL_PREFIX+"datasources."));
		String[] names =
			dir.list(GLUtils.acceptIfNameEndsWith(suf));
		if (names==null)
			names = new String[0];
		if (!all) {
			final int SUFN = suf.length();
			for (int i = names.length; --i >= 0;) {
				String nm = names[i];
				names[i] = nm.substring(0, nm.length()-SUFN);
			}
		}
		/*
		 * Unicode collation order only!
		 */
		Arrays.sort(names);
		return names;
	}

	/**
	 * Retrieves an error csw stored by the action page (via storeError).
	 */
	public CSWError
	getError() {
		if (getRequestParameters().containsKey(FORMERR_NAME)) {
			HttpSession ssn = getRequest().getSession(true);
			return (CSWError) ssn.getAttribute(FORMERR_NAME);
		}
		return null;
	}

	public String
	getFormValue(String key) {
		return getFormValue(this.getRequestParameters(), key);
	}
	/**
	 * Return the form value corresponding to KEY, or null if none.
	 * It's not enough just to say form.get(KEY) because
	 * <ul>
	 * <li>HTML &lt;input type='checkbox'> doesn't provide a key if the
	 * item wasn't checked.  GoLive writes a hidden input named ~KEY
	 * in this case.
	 * </li><li>HTML &lt;input type='image'> doesn't provide the item
	 * name directly, but it does provide KEY.x and KEY.y values
	 * corresponding to the coordinates of the click.
	 * </li></ul>
	 */
	public String
	getFormValue(Hashtable form, String key) {
		String val = null;
		if ((val = (String) form.get(key)) == null
		 && (val = (String) form.get("~"+key)) == null
		 && (val = (String) form.get(key+".x")) == null
		 && (val = (String) form.get(key+".y")) == null)
			;
		return val;
	}

	/**
	 * Returns the "friends" of this site as a ragged-right list of
	 * string lists.
	 * If A is the return value of getFriends(),
	 * and 0<=I<A.length, then A[I][0] is a string to be interpreted
	 * as a dotted-quad address; further, if 1<A[I].length,
	 * then A[I][1] is a string to be interpreted as a dotted-quad
	 * netmask.
	 */
	public String[][]
	getFriends()
		throws GLException
	{
		return GLTableReader.getData(this,
			new File(configFileName(GLUtils.GL_PREFIX+"friends")));
	}

	/**
	 * The output JspWriter for this page.
	 */
	public JspWriter
	getOut() {
		return pc.getOut();
	}

	/**
	 * The PageContext for this page.
	 */
	public PageContext
	getPageContext() {
		return pc;
	}

	/**
	 * Extracts a set of parameters from S.  These parameters are
	 * name/value pairs of strings; the return value is a hash
	 * table indexed by the parameter names.
	 * <p>
	 * getParams uses GLParamReader to read the parameters from an
	 * XML file.  S is a key meant to be useful to configFileName().
	 * <p>
	 */
	public Hashtable
	getParams(String s)
		throws GLException
	{
		String fname = configFileName(s);
		long mtime = new File(fname).lastModified();
		ServletContext appl = getApplication();
		Hashtable a = null;
		synchronized (appl) {
			a = (Hashtable) appl.getAttribute(s);
			/*
			 * If there's no data, or the file on disk is
			 * more recent than the data, we need to read
			 * the file.
			 */
			if (a == null || cacheMiss(mtime, a)) {
				a = GLParamReader.getData(this, fname);
				a.put(CACHE_MTIME, new Long(mtime));
				appl.setAttribute(s, a);
			}
		}
		return a;
	}

	/**
	 * Returns the request parameter named NAME, or DEFVAL
	 * if no such parameter.
	 */
	public String
	getParameter(String name, Object defval) {
		String val = (String) getRequestParameters().get(name);
		if (val != null)
			return val;
		if (defval == null || defval instanceof String)
			return (String) defval;
		return defval.toString();
	}

	public HttpServletRequest
	getRequest() {
		return (HttpServletRequest) pc.getRequest();
	}

	/**
	 * Convenience: returns the request parameters as a hash table.
	 * Currently this does nothing about character encoding of parameters;
	 * if the servlet API settles down, would be nice to reencode
	 * (e.g., request.setCharacterEncoding in Servlet API 2.3).
	 */
	public Hashtable
	getRequestParameters() {
		if (requestParameters==null)
			initRequestParameters();
		return requestParameters;
	}

	public HttpServletResponse
	getResponse() {
		return (HttpServletResponse) pc.getResponse();
	}

	/**
	 * Extracts a Locale from the HTTP request, and stores it in the
	 * request attributes so we compute it only once.
	 * Priority: (1) if user provided the magic $LC query parameter,
	 * use it; (2) otherwise use the "Accept-Language" header;
	 * (3) otherwise default en_US.
	 */
	public Locale
	getUserLocale()
	{
		HttpServletRequest req = getRequest();
		String ll = "", where = "";
		final String lname = GLUtils.GL_PREFIX + "locale";
		Locale lc = (Locale) req.getAttribute(lname);
		if (lc != null)
			return lc;
	set:
		{
			/* special flag */
			ll = req.getParameter(where="$LC");
			if (ll != null && (lc = glocale(ll)) != null)
				break set;
			/* Accept-Language: ja_JP */
			where = "getLocales()";
			ll = "[]";
			for (Enumeration e = req.getLocales();
			     e.hasMoreElements();)
			{
				lc = (Locale) e.nextElement();
				/* netscape bug: not fully locale-ized */
				if (Locale.ENGLISH.equals(lc))
					lc = Locale.US;
				break set;
			}
			/* fallback */
			where = "default";
			ll = "en_US";
			lc = Locale.US;
		}
		req.setAttribute(lname, lc);
		runtimeDebugMessage(new StringBuffer()
			.append(where).append(" locale is ").append(ll)
			.append(" == ").append(lc.getDisplayCountry())
			.append(" ").append(lc.getDisplayLanguage())
			.toString());
		return lc;
	}

	/**
	 * Persistent storage: adds O to the session attributes and
	 * returns the name under which we stored it for later retrieval.
	 */
	public String
	intern(Object o) {
		String n = internName();
		HttpSession ssn = getRequest().getSession(false);
		if (ssn != null)
			ssn.setAttribute(n, o);
		return n;
	}

	/**
	 * Generates URLs to navigate to next, previous, or arbitrary
	 * position in a record set.
	 */
	public String
	linkToARecord(ContentSourceWrapper csw, int i) {
		return linkToARecord(csw, i, false);
	}
	public String
	linkToARecord(ContentSourceWrapper csw, int i, boolean force) {
		/*
		 * Hmmm, hack.
		 */
		while (csw instanceof CSWBlockNavigatorView) {
			CSWBlockNavigatorView csn = (CSWBlockNavigatorView) csw;
			csw = csn.getViewedCSW();
			i = (i-1) * csw.BlockSize() + 1;
		}

		if (!force) {
			if (i < 1 || csw.RecordCount() < i
			 || csw.FirstRecordOnPage() == i)
				return "";
		}
		Hashtable a =
			resetRecordNavigation(getRequestParameters(), csw);
		a.put(GLUtils.rx(csw.GetName()), String.valueOf(i));
		return new StringBuffer()
			.append(" href='")
			.append(getRequest().getContextPath())
			.append(getRequest().getServletPath())
			.append('?')
			.append(GLUtils.queryArrayToString(a))
			.append("'")
			.toString();
	}
	public String
	linkToNewRecord(ContentSourceWrapper csw) {
		return new StringBuffer()
			.append(" href='")
			.append(HttpUtils.getRequestURL(
				(HttpServletRequest) pc.getRequest()))
			.append("?RECORD_KEY[")
			.append(csw.GetName())
			.append("]=$newRecord'")
			.toString();
	}
	public String
	linkToFirstRecord(ContentSourceWrapper csw) {
		return linkToARecord(csw, 1);
	}
	public String
	linkToLastRecord(ContentSourceWrapper csw) {
		final int nr = csw.RecordCount(),
			bs0 = csw.BlockSize(),
			bs = bs0 > 1 ? bs0 : 1,
			shift0 = nr%bs,
			shift = shift0 != 0 ? shift0 : bs;
		return linkToARecord(csw, nr-shift+1);
	}
	public String
	linkToNextRecord(ContentSourceWrapper csw) {
		return linkToARecord(csw,
			csw.FirstRecordOnPage() + csw.BlockSize());
	}
	public String
	linkToPreviousRecord(ContentSourceWrapper csw) {
		int m = csw.FirstRecordOnPage() - csw.BlockSize();
		return linkToARecord(csw, m > 1 ? m : 1);
	}

	/**
	 * Writes MESSAGE to the server log
	 * <p>tomcat's server log is logs/servlet.log.
	 */
	public void
	log(String message) {
		getApplication().log(
			new StringBuffer()
				.append(this.idstr)
				.append(' ')
				.append(message)
				.toString());
	}

	/**
	 * Converts P, a pathname presumed relative to our webapp,
	 * into a server-rooted pathname.
	 */
	public String mapPath(Object p) {
		String path = GLNull.nonnull(p).toString();
		/*
		 * Weed out paths that are already rooted, and things that
		 * look like URLs.
		 */
		if (path.indexOf("://") >= 0)
			return path;
		/*
		 * Hack around a bug in GoLive: prepend scheme://host:port,
		 * as well as our servlet context, to the URL.
		 */
		HttpServletRequest req = getRequest();
		String full = HttpUtils.getRequestURL(req).toString(),
			me = req.getContextPath();
		int ix = full.indexOf(me);
		if (ix < 0)
			return path;
		return new StringBuffer()
			.append(full.substring(0, ix))
			.append(path.startsWith("/") ? "" : me)
			.append(path.startsWith("/") ? "" : "/")
			.append(path)
			.toString();
	}

	/**
	 * Constructs and returns a hash table whose keys are the key[i]
	 * in KEY and whose values are obtained from the request parameters.
	 * KEY is a string of the form "key1,key2,key3..."
	 */
	public Hashtable
	mkkey(String key) {
		String[] keynames = GLUtils.split(",", key);
		Hashtable keyval = new Hashtable(keynames.length);
		ServletRequest req = getRequest();
		for (int j = 0; j < keynames.length; ++j) {
			keyval.put(keynames[j],
				GLNull.nonnull(
					req.getParameter(keynames[j]))
					.toString());
		}
		return keyval;
	}

	/**
	 * Moves the CSW's position off the end of the data.
	 * For the convenience of hand-coders; GoLive doesn't generate
	 * or recognize it.
	 */
	public void
	moveToEnd(ContentSourceWrapper csw) {
		csw.MoveTo(1+csw.RecordCount());
	}

	/**
	 * Returns the request parameter named NAME, or DEFVAL
	 * if no such parameter.
	 */
	public String
	pageParameter(String name, Object defval) {
		return getParameter(name, defval);
	}

	/**
	 * Sets a redirect header on the response.
	 * Expected usage is <pre>
	 *	redirect(target, args);
	 *	return;
	 * </pre>, or via the <code>gl:redirect</code> tag.
	 */
	public void
	redirect(String target, Hashtable args)
		throws GLException, IOException
	{
		if (args != null) {
			String as = args==null ? ""
				: GLUtils.queryArrayToString(args);
			if (!"".equals(as))
				target = target
					+ (target.indexOf('?')<0 ? '?' : '&')
					+ as;
		}

		/*
		 * In debug mode, forward to our reporting page (which
		 * in turn gives the user a chance to click through to
		 * the redirect).
		 */
		if (runtimeDebug) {
			try {
				getRequest()
					.setAttribute(REDIRECT_NAME, target);
				getPageContext().forward(
					"/config/actions/redebug.jsp");
			}
			catch (ServletException e) {
				throw new GLException(e.toString());
			}
		}
		else
			getResponse().sendRedirect(target);
	}

	/**
	 * Like the 2-argument version, but adds 1 bit more of debug info.
	 */
	public void
	redirect(String target, Hashtable args, boolean ok)
		throws GLException, IOException
	{
		getRequest().setAttribute(REDIRECT_OK_NAME,
			ok ? Boolean.TRUE : Boolean.FALSE);
		redirect(target, args);
	}

	/**
	 * Adds a function (since this is Java Jeopardy, it's in the form
	 * of an object) to be called by this.closeAll() on exit.
	 */
	public void
	registerCloser(GLNeedsCloser o) {
		needsClosing.add(o);
	}

	/**
	 * All content sources must register themselves so I can clean
	 * them up at the end of page processing.
	 */
	public void
	registerCSW(final ContentSourceWrapper csw) {
		registerCloser(
			new GLNeedsCloser() {
				public void close() {
					csw.Close();
				}
			});
	}

	/**
	 * Called by sensitive pages to make sure the user is an
	 * author or administrator of the web application.
	 */
	public void
	rejectUnauthorizedCallers()
		throws GLException
	{
		if (!friendly)	/* have not already checked this session */
			rejectUnauthorizedCallersViaFriends(getFriends());
		HttpSession ssn = getRequest().getSession(false);
		if (ssn != null)
			ssn.setAttribute(friendlyName, friendlyName);
		friendly = true;
	}

	/**
	 * Implement the security policy schematized in the FRIENDS array.
	 * We only check whether the requesting host matches one of the
	 * host.netmask pairs in the array.
	 */
	private void
	rejectUnauthorizedCallersViaFriends(String[][] friends)
		throws GLException
	{
		String h = "", h2 = "";
		final byte[] maskFF = { -127, -127, -127, -127 };
		ServletRequest req = pc.getRequest();
		boolean ok = false;
		try {
			byte[] reqaddr =
				InetAddress.getByName(h=h2=req.getRemoteAddr())
					.getAddress();
			runtimeDebugMessage("friends.length = "
				+ friends.length);
			for (int i = 0; i < friends.length; ++i) {
				byte[] okaddr =
					InetAddress.getByName(h=friends[i][0])
						.getAddress();
				byte[] mask = friends[i].length < 2 ? maskFF
					: InetAddress.getByName(h=friends[i][1])
						.getAddress();
				runtimeDebugMessage(
					GLUtils.cat("host ", istr(okaddr),
						" mask ", istr(mask)));
				if (eqaddr(reqaddr, okaddr, mask)) {
					ok = true;
					break;
				}
			}
		}
		catch (UnknownHostException e) {
			throw new GLException("Invalid hostname"
				+ h + "\n" + e.getMessage());
		}
		if (!ok)
			throw new GLException("I don't know you, " + h2, 403);
	}
	/**
	 * this.rejectUnauthorizedCallers() sets friendly TRUE on
	 * success, throws an exception on failure.  If we never
	 * call it, friendly is FALSE.
	 */
	public boolean friendly = false;

	/**
	 * Adds MESSAGE to a buffer maintained on a request attribute.
	 * Later code can retrieve the buffer and display it.
	 */
	public boolean runtimeDebug = false;
	public boolean runtimeErrorOccurred = false;
	public void
	runtimeDebugMessage(String message) {
		if (runtimeDebug)
			runtimeMessage(message);
	}
	/**
	 * Retrieves stored debugging messages.
	 */
	public StringBuffer
	runtimeDebugMessageBuffer() {
		final String name = GLUtils.GL_PREFIX + "DebugTrace";
		ServletRequest req = this.getRequest();
		StringBuffer bs = (StringBuffer) req.getAttribute(name);
		if (bs==null)
			req.setAttribute(name, bs=new StringBuffer());
		return bs;
	}
	public void
	runtimeErrorMessage(String message) {
		runtimeMessage(message);
	}
	private void
	runtimeMessage(String message) {
		StringBuffer bs = runtimeDebugMessageBuffer();
		bs.append("\n").append(message);
		log(message);
	}

	/**
	 * "selected"/"" as COND is true/false.
	 */
	public static String
	selected(boolean cond) {
		return cond ? "selected" : "";
	}

	/**
	 * Turns debug tracing on/off.
	 * For the whole webapp, but does not persist across restarts.
	 */
	public void
	setRuntimeDebug(boolean p) {
		final String pname = GLUtils.GL_PREFIX+"runtimeDebug";
		runtimeDebug = p;
		getApplication().setAttribute(pname, p ? "true" : "false");
	}

	/**
	 * For GoLive.  Produces an attribute suitable for marking an html
	 * label/span/td as containing an error.
	 */
	public String
	showError(Object o) {
		return GLCondition.HasError(this, o)
			? "class='error'" : "";
	}

	/**
	 * Puts an error object into the session.  Later, when we
	 * have redirected to a form page, the form page can grab
	 * the error object (via getError()) and use it to populate
	 * its form (hence can display error messages to the user).
	 */
	public void
	storeError(CSWError cse) {
		HttpSession ssn = getRequest().getSession(false);
		if (ssn != null)
			ssn.setAttribute(FORMERR_NAME, cse);
	}

	/**
	 * Determines and returns a truth value for O.
	 * I'm not sure I should give special treatment to String
	 * values like "true" or "false" -- O came directly from a
	 * database, so interpreting the strings would constitute
	 * another, gl-specific, filter on your bits.
	 */
	public static boolean
	truth(Object o) {
		if (GLNull.val.equals(o))
			return false;
		if (o instanceof java.math.BigDecimal)
			return ((java.math.BigDecimal) o).signum() != 0;
		if (o instanceof Number)
			return ((Number) o).doubleValue() != 0.0;
		if (o instanceof Boolean)
			return ((Boolean) o).booleanValue();
		if (o instanceof String)
			return !"".equals(o);
		return true;
	}

	/**
	 * Removes the attribute named N and returns its value.
	 */
	public Object
	unintern(String n) {
		HttpSession ssn = getRequest().getSession(false);
		if (ssn==null)
			return null;
		synchronized (ssn) {
			Object o = ssn==null ? null : ssn.getAttribute(n);
			if (o != null)
				ssn.removeAttribute(n);
			return o;
		}
	}

	/**
	 * Convenience: returns a string encoding the request parameters.
	 * This is a misnomer, since it captures all the request parameters,
	 * not just GET parameters from the URL.
	 */
	public String
	URLArgs() {
		return GLUtils.queryArrayToString(getRequestParameters());
	}

	/**
	 * Turns A, an array of strings like {"arg1","val1","arg2","val2"[,...]}
	 * into arg1=val1&arg2=val2[&...]
	 */
	public String
	URLArgs(Object[] a) {
		Hashtable h = new Hashtable();
		/* odd N induces exception */
		final int N = (a.length + 1) & ~1;
		for (int i = 0; i < N; i += 2)
			h.put(a[i], a[i+1]);
		return GLUtils.queryArrayToString(h);
	}

	/**
	 * Writes hidden parameters into the HTML.
	 * <p>A utility routine for Action setup methods.
	 */
	public void
	writeSetupHtml(Hashtable a)
	{
		try {
			JspWriter out = getOut();
			for (Enumeration e = a.keys(); e.hasMoreElements();) {
				String s = (String) e.nextElement();
				GLHOut.hiddenInput(out, s,
					GLNull.nonnull(a.get(s)).toString());
			}
			GLHOut.hiddenInput(out, "$CS",
				getResponse().getCharacterEncoding());
		}
		catch (Exception e) {
			runtimeErrorMessage(
				"writeSetupHtml took exception!");
		}
	}

	/**
	 * Set true if no RECORD_INDEX parameter.
	 */
	public boolean defaultedRecordIndex = false;
	/**
	 * For GLDebug.
	 */
	public Throwable exception = null;
	/**
	 * For profiling, &amp;c.
	 */
	public final long starttime;
	/**
	 * For logging, &amp;c.
	 */
	public static final long firststart = System.currentTimeMillis();
	public final int id;
	public final String idstr;
	public static final String idstrName =
		GLUtils.GL_PREFIX + "requestID";
	/**
	 * Use as key in pageContext.findAttribute() to retrieve gl.
	 */
	public static final String GL_NAME = "gl";
	public static final String JDBC_DSUF = ".jdbc.sbs";
	public static final String NL = System.getProperty("line.separator");

	public static final boolean SUCCESS = true, FAILURE = false;

	public static final String FORMERR_NAME = "FORMERR";
	public static final String REDIRECT_NAME = "$$redirect";
	public static final String REDIRECT_OK_NAME = "$$redirect-OK";

	/*
	 * -- Private methods
	 */

	private boolean
	cacheMiss(long mtime, Hashtable a) {
		final Long cmtime = (Long) a.get(CACHE_MTIME);
		return cmtime==null || cmtime.longValue() < mtime;
	}

	/*
	 * True iff two 32-bit internet addresses REQADDR and OKADDR
	 * are equivalent modulo MASK.
	 */
	private static boolean
	eqaddr(byte[] reqaddr, byte[] okaddr, byte[] mask) {
		for (int j = 0; j < mask.length; ++j)
			if ((reqaddr[j] & mask[j]) != (okaddr[j] & mask[j]))
				return false;
		return true;
	}

	/**
	 * Gets the boolean attribute named PNAME from application CTXT,
	 * backed by the application's init parameters, backed by DEFVAL.
	 */
	private static synchronized boolean
	getBooleanAttribute(ServletContext ctxt, String pname, boolean defval)
	{
		String pval = (String) ctxt.getAttribute(pname);
		if (pval == null) {
			pval = ctxt.getInitParameter(pname);
			boolean val = pval==null ? defval
				: "true".equalsIgnoreCase(pval);
			pval = val ? "true" : "false";
			ctxt.setAttribute(pname, pval);
		}
		return pval.equalsIgnoreCase("true");
	}

	private static synchronized boolean
	getDebug(ServletContext ctxt) {
		final String pname = GLUtils.GL_PREFIX+"runtimeDebug";
		String pval = (String) ctxt.getAttribute(pname);
		if (pval == null) {
			pval = ctxt.getInitParameter(pname);
			if (!"true".equalsIgnoreCase(pval))
				ctxt.setAttribute(pname, pval = "false");
		}
		return pval.equalsIgnoreCase("true");
	}

	private static boolean
	getDoRecode(ServletContext ctxt) {
		final String pname = GLUtils.GL_PREFIX+"doRecode";
		return getBooleanAttribute(ctxt, pname, true);
	}

	/**
	 * Returns the java.util.Locale that LL names.
	 * <p>LL looks like "ja_JP" (language_COUNTRY) or perhaps "ja"
	 * (language only).
	 */
	private Locale
	glocale(String ll) {
		String[] ls = GLUtils.split("_", ll);
		return new Locale(ls[0], ls.length<2?"":ls[1]);
	}

	/**
	 * Sets up the internal requestParameters hash for use by
	 * getRequestParameters, &amp;c.
	 * <p>
	 * For want of a better strategy, reencodes request parameters
	 * (1) according to a "$CS" request parameter, if passed
	 * (GoLive-created pages do this automatically)
	 * (2) according to its own output character encoding.
	 * <p>
	 * It would be nice to use request.setCharacterEncoding, but
	 * that is Servlet API 2.3, and it has its own problems.
	 */
	private void
	initRequestParameters() {
		ServletRequest req = pc.getRequest();
		ServletResponse resp = pc.getResponse();
		String newCharset = req.getParameter("$CS");
		if (newCharset==null)
			newCharset = resp.getCharacterEncoding();
		Hashtable a = new Hashtable();
		for (Enumeration e = req.getParameterNames();
			e.hasMoreElements();)
		{
			final String name = (String) e.nextElement();
			String value = req.getParameter(name);
			if (doRecode) {
				value = recode(value, newCharset);
			}
			a.put(name, value);
		}
		requestParameters = a;
	}

	/**
	 * Generates a unique name for intern() to use.
	 */
	private String
	internName() {
		String t = GLNumberFormat.Number(
			this, new Integer(++internID), 0, 0, 3, 0);
		return idstr+'#'+t;
	}

	/**
	 * Converts A into a dotted-decimal Internet address string.
	 */
	private static String
	istr(byte[] a) {
		StringBuffer ts = new StringBuffer((3+1)*a.length-1);
		String dot = "";
		for (int i = 0; i < a.length; ++i, dot = ".")
			ts.append(dot).append(a[i]&0xff);
		return ts.toString();
	}

	/**
	 * Identifier for GLContext objects.
	 */
	private int
	nextID() {
		ServletContext appl = getApplication();
		int nextid = 0;
		synchronized (appl) {
			try {
				Integer sid =
					(Integer) appl.getAttribute(IDGEN_NAME);
				nextid = sid.intValue();
			}
			catch (Throwable e) {
			}
			appl.setAttribute(IDGEN_NAME, new Integer(++nextid));
		}
		return nextid;
	}

	/**
	 * Converts S, which is understood to be a sequence of byte
	 * values stored as a String, into an actual Unicode string
	 * using ENC as the name of the character encoding to use.
	 */
	private String
	recode(String s, String enc) {
		if (s==null || "".equals(s))
			return s;
		byte []b = new byte[s.length()];
		for (int i = 0; i < b.length; ++i)
			b[i] = (byte) (s.charAt(i) & 0xff);
		try {
			s = new String(b, enc);
		}
		catch (Exception e) {
			runtimeDebugMessage(e.toString());
		}
		return s;
	}

	/**
	 * Returns a copy of A with RECORD_{KEY,INDEX}[cswname],
	 * and all of the primary key arguments, stripped out.
	 */
	private Hashtable
	resetRecordNavigation(Hashtable a, ContentSourceWrapper csw) {
		Hashtable na = new Hashtable();
		String[] pkey = csw.PrimaryKey();
		Hashtable pkh = new Hashtable();
		for (int i = 0; i < pkey.length; ++i)
			pkh.put(pkey[i], pkey[i]);

		String rk = GLUtils.rk(csw.GetName()),
			rx = GLUtils.rx(csw.GetName());
		for (Enumeration e = a.keys(); e.hasMoreElements();) {
			String key = (String) e.nextElement();
			if (!rk.equals(key) && !rx.equals(key)
			 && pkh.get(key)==null)
				na.put(key, a.get(key));
		}
		return na;
	}

	private PageContext pc;
	private Hashtable requestParameters;
	private Vector needsClosing = new Vector();
	private int internID;
	private String aspec;
	private Vector avec;
	private boolean doRecode = true;

	private static final String CACHE_MTIME = "  #mtime  ";
	private static final String friendlyName =
		GLUtils.GL_PREFIX + "friendly";
	private static final String FSP = File.separator;
	private static final String IDGEN_NAME =
		GLUtils.GL_PREFIX + "idgen";
	/**
	 * Converts between abstract keys and names in URL space relative
	 * to the webapp root.  We depend on URL space and the file system
	 * having the same layout below this point.
	 *
	 * @see #configFileName
	 */
	private static final Hashtable cfmap;
	static {
		cfmap = new Hashtable();
		cfmap.put(GLUtils.GL_PREFIX + "datasources.",
			"config"+FSP+"datasources"+FSP+"?"+JDBC_DSUF);
		cfmap.put(GLUtils.GL_PREFIX + "friends",
			"WEB-INF"+FSP+"friends.xml");
		cfmap.put(GLUtils.GL_PREFIX + "ui-info.",
			"config"+FSP+"info"+FSP+"?.ui6.xml");
	}
};
