/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

import java.io.*;
import java.lang.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import javax.servlet.jsp.*;

/**
 * Error/debug report.
 * Calling the static <code>getInstance()</code> method prepares a report
 * containing information about therequest, session, thrown exception
 * (if applicable), and any tracing information written by
 * gl.runtime{Debug,Error}Message() during this request.
 * Calling the instance's <code>toString()</code> method
 * retrieves the report.
 *
 * <p>
 * The GLDebug instance also contains a number of state flags that can
 * be used by error pages.  These are represented as
 * <strong>public final</strong> instance variables.
 * (This line added to work around an InstallSheild checksum bug)
 */
public class GLDebug {
	/*
	 * -- Constructors.
	 */
	private GLDebug(GLContext gl) {
		this(gl, gl.exception);
	}
	private GLDebug(GLContext gl, Throwable exception) {
		HttpServletRequest request = gl.getRequest();

		if (b)
			throw new NullPointerException("GLDebug");
		err = exception==null ? 0
			: !(exception instanceof GLException) ? 500
			: ((GLException) exception).getHttpError();
		advise = err >= 500;
		log = exception != null || gl.runtimeErrorOccurred;
		trace = gl.runtimeDebug && gl.friendly;
		admin_email = !advise ? ""
			: GLNull.nonnull(
				gl.getApplication()
					.getInitParameter(GLUtils.GL_PREFIX
						+ "admin_email")
				).toString();

		if (log || trace) {
			ts.append("<h4>Exception</h4>").append(NL);
			if (exception == null)
				ts.append("none.").append(NL);
			else
				reportException(exception);
			reportRequest(request);
			reportRequestParameters(request);
			reportRequestHeaders(request);
			reportRequestAttributes(request);
			/*
			 * Log it before adding the trace information
			 * (since that's already in the log).
			 */
			if (log)
				gl.log(ts.toString());
			/* Cookies are already in headers */
			/* reportCookies(request); */
			reportTrace(gl);
		}

		logName = gl.idstr;
	}
	/*
	 * -- Public.
	 */
	public static GLDebug
	getInstance(GLContext gl) {
		final String name = GLUtils.GL_PREFIX + "GLDebug";
		HttpServletRequest req = gl.getRequest();
		GLDebug val = (GLDebug) req.getAttribute(name);
		if (val==null)
			req.setAttribute(name, val=new GLDebug(gl));
		return val;
	}
	public String toString() {
		return ts.toString();
	}
	/**
	 * The error that occurred, if it can be represented as an
	 * HTTP error code; or 0, if no error actually occurred.
	 */
	public final int err;
	/**
	 * True if a log was generated.
	 * This log isn't accessible via HTTP, but a site administrator
	 * can inspect it.
	 */
	public final boolean log;
	/**
	 * Log identifier, if LOG is true.
	 * We use this to tag entries in the servlet log.
	 */
	public final String logName;
	/**
	 * True if the error page include trace information in its
	 * output.  The standard /WEB-INF/jspi/footer.jspi uses this.
	 */
	public final boolean trace;
	/**
	 * True if the error page should include a message telling the
	 * user to advise the site administrator of the problem.
	 * This is pretty simpleminded: error 500 generates advice, but
	 * others don't.
	 */
	public final boolean advise;
	/**
	 * Email address for the site administrator.
	 */
	public final String admin_email;

	/*
	 * -- Private.
	 */

	/**
	 * Reports on cookies.
	 * XXX since these are already in the headers, just leave them out?
	 */
	private void
	reportCookies(HttpServletRequest req) {
		Cookie[] a = req.getCookies();
		if (a==null)
			return;
		ts.append("<h4>Cookies:</h4>").append(NL);
		for (int i = 0; i < a.length; ++i) {
			if (a[i]==null)
				return;
			ts.append("Cookie ").append(a[i].getName())
				.append(" = ").append(a[i].getValue())
				.append(NL);
			ts.append(NL);
		}
	}

	private void
	reportException(Throwable e) {
		reportException(e, ts);
	}

	public static void
	reportException(Throwable e, StringBuffer ts) {
		StringWriter writer = new StringWriter();
		e.printStackTrace(new PrintWriter(writer));
		ts.append(writer.getBuffer());
		ts.append(NL);
	}

	private void
	reportRequest(HttpServletRequest req) {
		ts.append("<h4>Request: </h4>");
		ts.append(req.getMethod());
		ts.append(" ");
		ts.append(HttpUtils.getRequestURL(req));
		String q = req.getQueryString();
		if (q!=null && q.length()>0) {
			ts.append("?");
			ts.append(q);
		}
		ts.append(NL);
		{
			String ssn = req.getRequestedSessionId();
			ts.append("<h4>Session: </h4>");
			if (ssn==null)
				ts.append("none");
			else if (!req.isRequestedSessionIdValid())
				ts.append("invalid");
			else
				ts.append(ssn);
		}
	}

	private void
	reportRequestHeaders(HttpServletRequest req) {
		String p = NL, none = " none";
		ts.append("<h4>Request Headers:</h4>");
		ts.append(htmlTableStartTag);
		for (Enumeration e = req.getHeaderNames();
			e != null && e.hasMoreElements();)
		{
			String h = (String) e.nextElement();
			ts.append("<tr><td>");
			p = ""; none = "";
			ts.append(h).append("</td>\r\n<td>")
				.append(req.getHeader(h))
				.append("</td></tr>\r\n");

		}
		ts.append("</table>\r\n");
		ts.append(none);
	}

	private void
	reportRequestParameters(HttpServletRequest req) {
		ts.append("<h4>Request Parameters:</h4>");
		String p = NL, none = "<tr><td>none</td></tr>\r\n";
		ts.append(htmlTableStartTag);
		for (Enumeration e = req.getParameterNames();
			e.hasMoreElements();)
		{
			p = ""; none = "";
			String name = (String) e.nextElement();
			String[] values =
				req.getParameterValues(name);
			for (int i = 0; i < values.length; ++i) {
				ts.append("<tr><td>")
					.append(name).append("</td>\r\n<td>")
					.append(values[i])
					.append("</td></tr>\r\n");
				name = "+ ";
			}
		}
		ts.append(none);
		ts.append("</table>\r\n");
		ts.append(NL).append("<h4>Character Encoding: </h4>")
			.append(req.getCharacterEncoding())
			.append(NL);
	}

	private void
	reportRequestAttributes(HttpServletRequest req) {
		ts.append("<h4>Request Attributes:</h4>");
		String p = NL, none = " none";
		final String tracename = GLUtils.GL_PREFIX + "DebugTrace";
		ts.append(htmlTableStartTag);
		for (Enumeration e = req.getAttributeNames();
			e.hasMoreElements();)
		{
			String name = (String) e.nextElement();
			if (tracename.equals(name))
				continue;
			ts.append("<tr><td>");
			none = "";
			ts.append(name).append("</td>\r\n<td>")
				.append(GLNull.nonnull(req.getAttribute(name))
					.toString())
				.append("</td></tr>\r\n");
		}
		ts.append("</table>\r\n");
		ts.append(none);
	}

	private void
	reportTrace(GLContext gl)
	{
		StringBuffer bs = gl.runtimeDebugMessageBuffer();
		ts.append("<h4>Trace Data:</h4>");
		if (bs.length()==0)
			ts.append(" none.");
		else {
			ts.append(NL);
			ts.append(bs);
		}
		ts.append(NL);
	}
	private StringBuffer ts = new StringBuffer();
	private boolean b = false;

	private static final String htmlTableStartTag =
		"<table border='1' cellspacing='2' cellpadding='1'>";
	private static final String NL = "\r\n";
};
