/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.utils;

import javax.servlet.*;
import javax.servlet.http.*;
import javax.servlet.jsp.*;
import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;

/**
 * GLUtils is a grab-bag of utilities, declared static so they don't
 * need a this object.
 */
public class GLUtils {
	/**
	 * Constructor is private because I don't think this ever needs
	 * to be instantiated.
	 */
	private GLUtils() {
	}

	/**
	 * Returns a FilenameFilter instance, useful in
	 * <code>java.io.File.list</code>,
	 * that accepts files whose names end in SUFFIX.
	 */
	public static FilenameFilter
	acceptIfNameEndsWith(final String suffix) {
		return new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(suffix);
			}
		};
	}

	/**
	 * Returns '$'ACTION'_'FLAG, useful for embedding parameters
	 * pertinent to specific GL actions into HTML forms.
	 */
	public static String
	actionParam(String action, String flag)
	{
		return cat("$", action, "_", flag);
	}

	/**
	 * True iff at least one of the elements of A matches KEYSTR.
	 */
	public static boolean
	any(String[] a, String keystr) {
		for (int i = a.length; --i >= 0;)
			if (a[i].equalsIgnoreCase(keystr))
				return true;
		return false;
	}

	/**
	 * Returns YAT+YEE+SAHM+SEI.
	 * <p>Makes only one StringBuffer for the whole operation,
	 * instead of one per string concatenation (Java compilers
	 * maybe now can do better?)
	 * <p>You can certainly do this yourself:<pre>
	 * new StringBuffer().append(YAT).append(YEE).append(SAHM)
	 *	.append(SEI).toString();
	 * </pre>
	 */
	public static String
	cat(String yat, String yee, String sahm, String sei) {
		final int n = yat.length() + yee.length() + sahm.length()
			+ sei.length();
		return new StringBuffer(n)
			.append(yat).append(yee).append(sahm).append(sei)
			.toString();
	}

	/**
	 * Utility method always throws a GLException, hopefully with
	 * a helpful message about the ContentSourceWrapper O that the
	 * caller was hoping to process.
	 */
	public static void
	cswQuack(String cname, Object o)
		throws GLException
	{
		Object oname = o==null ? "(null)"
			: o instanceof String ? o
			: o instanceof ContentSourceWrapper
				? ((ContentSourceWrapper) o).GetName()
			: o.getClass().getName();
		throw new GLException(cname + ": invalid content source: "
			+ oname);
	}

	/**
	 * Dumps the input named by RR to the output named by OUT.
	 * Canonicalizes line feeds along the way (produces \r\n).
	 */
	public static void
	dumpTextFile(Reader rr, JspWriter out)
		throws IOException
	{
		for (;;) {
			int c = rr.read();
			if (c=='\n' || c=='\r' && (c=rr.read())=='\n') {
				out.write('\r');
				out.write('\n');
			}
			else if (c >= 0)
				out.write(c);
			else
				break;
		}
	}

	/**
	 * True iff each key mentioned in KEYSTR ("key1&amp;key2&amp;...")
	 * has a value in A.
	 */
	public static boolean
	has(Hashtable a, String keystr) {
		String[] keys = split("&", keystr);
		for (int i = 0; i < keys.length; ++i)
			if (a.get(keys[i]) == null)
				return false;
		return true;
	}

	/**
	 * Returns the keys of H as an array of strings.
	 * <p>If the keys aren't actually strings, we'll fall down, go boom.
	 */
	public static String[]
	hkeys(Hashtable h) {
		int i = 0;
		for (Enumeration e = h.keys(); e.hasMoreElements(); ++i) {
			String s = (String) e.nextElement();
		}
		String[] keynames = new String[i];
		i = 0;
		for (Enumeration e = h.keys(); e.hasMoreElements(); ++i)
			keynames[i] = (String) e.nextElement();
		return keynames;
	}

	/**
	 * Inserts HTML breaks (&lt;br>) between newlines.
	 */
	public static String
	htmlBreaks(Object os) {
		String s = GLNull.nonnull(os).toString();
		final int NS = s.length();
		StringBuffer ts = new StringBuffer();
		for (int i = 0; i < NS; ++i) {
			final char c = s.charAt(i);
			/*if (c==' ')
				ts.append("&nbsp;");
			else */if (c=='\n')
				ts.append("<br>\n");
			else if (c!='\r')
				ts.append(c);
		}
		return ts.toString();
	}

	/**
	 * Turns HTML magic characters into numeric character references.
	 * This protects them from being interpreted as HTML.
	 */
	public static String
	htmlEncode(Object os) {
		String s = GLNull.nonnull(os).toString();
		final int NS = s.length();

		/* See if any chars need conversion. */
	findhtchar:
		{
			for (int i = htmlc.length(); --i >= 0;)
				if (s.indexOf(htmlc.charAt(i)) >= 0)
					break findhtchar;
			return s;
		}

		StringBuffer w = new StringBuffer();
		for (int i = 0; i < NS; ++i) {
			final char c = s.charAt(i);
			if (htmlc.indexOf(c) < 0)
				w.append(c);
			else
				w.append("&#").append((int) c).append(";");
		}
		return w.toString();
	}

	/**
	 * Joins the strings in A using separator S.
	 */
	public static String
	join(String s, String[] a) {
		return join(s, a, 0, a.length);
	}

	/**
	 * Joins the strings in A[START:END) using separator S.
	 */
	public static String
	join(String s, String[] a, int start, int end) {
		String sep = "";
		StringBuffer w = new StringBuffer();
		if (end > a.length)
			end = a.length;
		for (int i = start; i < end; ++i, sep = s)
			w.append(sep).append(a[i]);
		return w.toString();
	}

	/**
	 * Puts the string DATA into file F, creating if necessary.
	 */
	public static int
	putFile(File f, String data)
	{
		OutputStream os = null;
		PrintWriter pr = null;
		int rc = 0;
		try {
			os = new FileOutputStream(f);
			pr = new PrintWriter(os);
			pr.print(data);
		}
		catch (NullPointerException e) {
			System.out.println(e.toString());
			rc = 400;
		}
		catch (Exception e) {
			System.out.println(e.toString());
			rc = 403;
		}
		finally {
			if (pr != null)
				try { pr.close(); } catch (Exception e) {}
			if (os != null)
				try { os.close(); } catch (Exception e) {}
		}
		return rc;
	}

	/**
	 * Formats A's contents as a string
	 * <code>key1=val1&amp;key2=val2&amp;...</code>
	 * and returns it.
	 */
	public static String
	queryArrayToString(Hashtable a) {
		Enumeration e = a.keys();
		StringBuffer t = new StringBuffer();
		String sep = "";
		while (e.hasMoreElements()) {
			String k = (String) e.nextElement();
			t.append(sep); sep = "&";
			t.append(URLEncoder.encode(k));
			t.append("=");
			t.append(URLEncoder.encode((String) a.get(k)));
		}
		return t.toString();
	}

	/**
	 * Formats A's contents as a string
	 * <code>key1=val1&amp;key2=val2&amp;...</code>
	 * and returns it.
	 * <p>Here A is an array of key-value pairs like
	 * <code>[ key1, val1, key2, val2, ... ]</code>
	 */
	public static String
	queryArrayToString(Object[] a) {
		/* odd length induces exception */
		final int N = (a.length + 1) & ~1;
		/* a little too big, but shouldn't matter */
		Hashtable h = new Hashtable(N);
		for (int i = 0; i < N; i+=2)
			h.put((String) a[i], a[i+1]);
		return queryArrayToString(h);
	}

	/**
	 * Extracts RECORD_INDEX[cswName] as an integer and returns it;
	 * or 1 if not found.
	 */
	public static int
	queryIndex(Hashtable query, String cswName) {
		String s = (String) query.get(rx(cswName));
		return s==null ? 1 : Integer.parseInt(s);
	}

	/**
	 * Returns a new string with instances of OS in SRC replaced by NS.
	 */
	public static String
	replace(String src, String os, String ns) {
		return join(ns, split(os, src));
	}

	public static String
	reqstr(HttpServletRequest req) {
		StringBuffer ts = new StringBuffer(req.getRequestURI());
		String q = req.getQueryString();
		if (q != null)
			ts.append('?').append(q);
		return ts.toString();
	}

	public static String
	rk(String cname) {
		return cat("RECORD_KEY[", cname, "]", "");
	}

	public static String
	rx(String cname) {
		return cat("RECORD_INDEX[", cname, "]", "");
	}

	/**
	 * Splits VAL into an array of substrings using separator SEP.
	 * <code>
	 * split("&amp;", "a=b&amp;x=y") => ["a=b", "x=y"];
	 * split("", "abc") => [ "a", "b", "c" ]
	 * </code>
	 */
	public static String []
	split(String sep, String val) {
		if ("".equals(sep)) {
			String[] a = new String[val.length()];
			for (int i = 0; i < a.length; ++i)
				a[i] = val.substring(i, i+1);
			return a;
		}
		Vector v = new Vector();
		final int vn = val.length(), sepn = sep.length();
		int vi = 0, vx;
		do {
			if ((vx = val.indexOf(sep, vi)) < 0)
				vx = vn;
			v.add(val.substring(vi, vx));
		} while ((vi = vx+sepn) < vn);
		String[] a = new String[v.size()];
		v.copyInto(a);
		return a;
	}

	/**
	 * Doubles up embedded single quotes in a string so that
	 * we can hand it to an SQL interpreter.
	 */
	public static String
	sqlEncode(String s) {
		if (s.indexOf("'") < 0)
			return s;
		String[] a = split("'", s);
		return join("''", a);
	}

	/**
	 * Shorthand for <code>x(s, n, "")</code>.
	 */
	public static String
	x(String s, int n) {
		return x(s, n, "");
	}

	/**
	 * Replicates S N times into a return string, using separator SEP.
	 * <pre>
	 * x("abc", 0) -> ""
	 * x("abc", 1) -> "abc"
	 * x("abc", 2) -> "abcabc"
	 * x("abc", 2, "+") -> "abc+abc+abc"
	 * </pre>
	 */
	public static String
	x(String s, int n, String sep) {
		if (n < 1)
			return "";
		String sp = "";
		StringBuffer ts =
			new StringBuffer(n*s.length() + (n-1)*sep.length());
		for (; --n >= 0; sp = sep)
			ts.append(sp).append(s);
		return ts.toString();
	}

	/**
	 * Prefix for application init parameters, etc.
	 */
	public static final String GL_PREFIX = "com.adobe.gl.";

	/**
	 * The Internet line terminator: "\r\n"
	 */
	public static final String CRLF = "\r\n";

	/**
	 * Characters that HTML thinks are magic.
	 * XML is almost the same but needs to add &gt;.
	 */
	private static final String htmlc = "&<\"'";
};
