/*
 * GoLive
 *
 * ADOBE SYSTEMS INCORPORATED
 * Copyright 2000-2002 Adobe Systems Incorporated. All Rights Reserved.
 * 
 * NOTICE:  Notwithstanding the terms of the Adobe GoLive End User 
 * License Agreement, Adobe permits you to reproduce and distribute this 
 * file only as an integrated part of a web site created with Adobe 
 * GoLive software and only for the purpose of enabling your client to 
 * display their web site. All other terms of the Adobe license 
 * agreement remain in effect.
 */

package com.adobe.gl.runtime6.xml;

import java.io.*;
import java.lang.*;
import java.net.*;
import java.util.*;
import com.adobe.gl.runtime6.utils.*;

import org.w3c.dom.*;
import org.w3c.dom.traversal.*;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import org.apache.xalan.xslt.*;
import org.apache.xalan.serialize.*;
import org.apache.xalan.templates.OutputProperties;
import org.apache.xpath.XPathAPI;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import org.xml.sax.EntityResolver;

/**
 * Utilities for XML parsing and transformation.
 * This class provides no instances, just a namespace.
 * Its methods try to rationalize exceptions thrown
 * by implementation classes.
 */
public class GLXML {

	private GLXML() {}

	/**
	 * Returns XP1+"/"+XP2, or XP1 if XP2 is empty, or XP2 if XP1
	 * is empty, or XP2 if XP2 is rooted (starts with "/").
	 * <p>I'm not sure the rooted case is the right thing.
	 */
	public static String
	addPath(String xp1, String xp2) {
		return "".equals(xp2) ? xp1
			: "".equals(xp1) || xp2.startsWith("/") ? xp2
			: xp1 + "/" + xp2;
	}

	/**
	 * Converts XML into a Document.
	 */
	public static Document
	asDom(GLContext gl, String xml)
		throws GLException, java.io.FileNotFoundException
	{
		xml = gl.contextPathFrom(xml);
		String idxml = xml;
		java.io.InputStream sxml = null;
		if (xml.startsWith("/")) {
			sxml = gl.getApplication().getResourceAsStream(xml);
			if (sxml==null)
				throw new FileNotFoundException("url "+xml);
			/*
			 * Create a special "context-relative" URL.
			 * Our custom entity resolver in setEntityResolver()
			 * knows how to process these.
			 */
			idxml = CX+xml;
		}
		/*
		 * My reading of the API is that we're not guaranteed a
		 * file:///-ready URL connection.  So I hack a special
		 * case here, one perhaps too forgiving of mistyped URLs.
		 */
		else if (xml.startsWith("file:")) {
			int N = xml.length(), i = "file:".length();
			for (; i < N; ++i)
				if (xml.charAt(i)!='/'
				 && xml.charAt(i)!=File.separatorChar)
					break;
			if (File.separatorChar=='/')
				--i;
			String a = xml.substring(i);
			sxml = new FileInputStream(new File(a));
			idxml = a;
		}
		else {
			sxml = GLHttp.asStream(xml);
		}

		InputSource in = null;
		if (!debug())
			in = new InputSource(sxml);
		else {
			System.out.println("DUMP "+xml+" BEGIN");
			InputStream mxml = dupInputStream(sxml);
			System.out.println("DUMP "+xml+" END");
			in = new InputSource(mxml);
		}
		gl.runtimeDebugMessage("in.setSystemId "+idxml);
		in.setSystemId(idxml);

		Document doc = null;
		try {
			doc = asDom(gl, in);
		}
		finally {
			try { sxml.close(); } catch (Exception e) {}
		}
		return doc;
	}

	/**
	 * Converts XML stored in XFILE into a Document.
	 */
	public static Document
	asDom(GLContext gl, File xfile)
		throws GLException, java.io.FileNotFoundException
	{
		/* "I want to believe." */
		InputStream sxml = new FileInputStream(xfile);
		InputSource in = new InputSource(sxml);
		String idxml = xfile.getAbsolutePath();
		gl.runtimeDebugMessage("in.setSystemId "+idxml);
		in.setSystemId(idxml);
		Document doc = null;
		try {
			doc = asDom(gl, in);
		}
		finally {
			try { sxml.close(); } catch (Exception e) {}
		}
		return doc;
	}

	/**
	 * Converts XML read from IN into a Document.
	 */
	public static Document
	asDom(GLContext gl, InputSource in)
		throws GLException
	{
		DocumentBuilderFactory dfactory =
			DocumentBuilderFactory.newInstance();
		dfactory.setNamespaceAware(true);
		dfactory.setValidating(false);
		DocumentBuilder dbuilder = null;
		Document doc = null;
		try {
			dbuilder = dfactory.newDocumentBuilder();
			setEntityResolver(gl, dbuilder);
			doc = dbuilder.parse(in);
		}
		catch (Exception e) {
			throw new GLException(e.toString());
		}
		return doc;
	}

	/**
	 * Returns a string representation of the nodes below N0
	 * selected by XPATH and a default Transformer.
	 */
	public static String
	asText(Node n0, String xpath)
		throws GLException
	{
		try {
			return asText(n0, xpath, serializer());
		}
		catch (TransformerConfigurationException e) {
			throw new GLException(e.toString());
		}
	}

	/**
	 * Returns a string representation of the nodes below N0
	 * selected by XPATH and the caller's Transformer.
	 */
	public static String
	asText(Node n0, String xpath, Transformer serializer)
		throws GLException
	{
		StringBuffer ts = new StringBuffer();
		try {
			NodeIterator nl = selectNodeIterator(n0, xpath);
			Node n;
			while ((n = nl.nextNode()) != null) {
				java.io.StringWriter wr =
					new java.io.StringWriter();
				serializer.transform(
					new DOMSource(n),
					new StreamResult(wr));
				StringBuffer a = wr.getBuffer();
				ts.append(a.toString());
			}
		}
		catch (Exception e) {
			throw new GLException(e.toString());
		}
		return ts.toString();
	}

	private static boolean
	debug() {
		return false;
	}

	/**
	 * For debugging -- dumps the tree.
	 */
	public static void
	dumpDom(Node topnode, java.io.OutputStream out) {
		try {
			Serializer ser =
				SerializerFactory.getSerializer(
					OutputProperties
					.getDefaultMethodProperties("xml"));
			ser.setOutputStream(System.out);
			ser.asDOMSerializer().serialize(topnode);
		}
		catch (Exception e) {
			/* oh, well */
		}
	}

	/**
	 * Returns the list of nodes selected by XPATH at context NODE.
	 * The list might be empty, so it can return a 0-length array.
	 */
	public static Node[]
	getNodes(Node node, String xpath)
		throws GLException
	{
		Vector v = null;
		NodeIterator nl = selectNodeIterator(node, xpath);
		Node n;
		while ((n = nl.nextNode()) != null) {
			if (v==null) v = new Vector();
			v.add(n);
		}
		nl = null;
		if (v==null)	/* a common case */
			return nonodes;
		Node[] a = new Node[v.size()];
		v.copyInto(a);
		return a;
	}

	/**
	 * Returns a NodeIterator interface object, hiding details like
	 * the implementation class's name and exceptions.
	 */
	public static NodeIterator
	selectNodeIterator(Node n0, String xpath)
		throws GLException
	{
		try {
			return XPathAPI.selectNodeIterator(n0, xpath);
		}
		catch (Exception e) {
			throw new GLException(e.toString());
		}
	}

	/**
	 * Produces a Transformer for serializing
	 * XML output to character/byte streams.
	 * <p>
	 * Sets the <code>omit-xml-declaration</code>
	 * property true, so output won't have the
	 * declaration.
	 */
	public static Transformer
	serializer()
		throws TransformerConfigurationException
	{
		Transformer ser = TransformerFactory.newInstance()
			.newTransformer();
		ser.setOutputProperty(
			"omit-xml-declaration", "yes");
		return ser;
	}

	/**
	 * Constructs a custom entity resolver, and attaches it to DBUILDER.
	 * During parsing, DBUILDER will give this resolver first crack
	 * at any entities it comes across.
	 * <p>
	 * If this resolver sees a DTD reference, it produces an empty
	 * byte stream.
	 * If it sees a special context-relative URL, it resolves relative
	 * to the running web application.
	 */
	private static void
	setEntityResolver(final GLContext gl, DocumentBuilder dbuilder) {
		EntityResolver resolver = new EntityResolver() {
			public InputSource
			resolveEntity(String pub, String sys) {
				InputSource es = null;
				gl.runtimeDebugMessage(
					"resolveEntity pub "+pub+" sys "+sys);
				/*
				 * Just throw away the DTD, since we're not validating.
				 */
				if (sys!=null
				 && sys.toLowerCase().endsWith(".dtd"))
				{
					es = new InputSource(
						new ByteArrayInputStream(
							new byte[0]));
					gl.runtimeDebugMessage(
						"Threw away DTD "+sys);
				}
				/*
				 * Special context-relative URLs.
				 */
				else if (sys!=null && sys.startsWith(CX)) {
					String tsys = sys.substring(CX.length());
					java.io.InputStream sxml =
						gl.getApplication()
							.getResourceAsStream(tsys);
					if (sxml!=null)
						es = new InputSource(sxml);
				}

				/*
				 * We have an input source.  Set its public and system
				 * IDs appropriately (in case it references other
				 * entities).
				 */
				if (es!=null) {
					if (pub!=null)
						es.setPublicId(pub);
					if (sys!=null)
						es.setSystemId(sys);
					return es;
				}
				/*
				 * NULL return means appeal to the default behavior.
				 */
				return null;
			}
		};
		dbuilder.setEntityResolver(resolver);
	}

	/**
	 * Transforms DOC according to SDOC;
	 * returns the new transformed Document.
	 */
	public static Document
	transform(String xml, Document doc, Document sdoc)
		throws GLException
	{
		try {
			TransformerFactory tfactory =
				TransformerFactory.newInstance();
			DOMSource dxsl = new DOMSource(sdoc);
			Transformer transformer =
				tfactory.newTransformer(dxsl);
			DOMSource dxml = new DOMSource(doc);
			dxml.setSystemId(xml);
			DOMResult dxres = new DOMResult();
			transformer.transform(dxml, dxres);
			return (Document) dxres.getNode();
		}
		catch (Exception e) {
			throw new GLException(e.toString());
		}
	}

	/**
	 * Returns a copy of IS, having dumped the contents of IS to
	 * System.out as a debugging message.
	 */
	private static InputStream
	dupInputStream(InputStream is)
		throws GLException
	{
		try {
			ByteArrayOutputStream bs = new ByteArrayOutputStream();
			int c;
			while ((c = is.read()) >= 0)
				bs.write(c);
			byte[] b = bs.toByteArray();
			System.out.write(b);
			return new ByteArrayInputStream(b);
		}
		catch (IOException e) {
			e.printStackTrace(System.out);
			throw new GLException(e.toString());
		}
	}

	static private final Node[] nonodes = new Node[0];

	/**
	 * Scheme for special context-relative URLs.
	 * @see #setEntityResolver
	 */
	static private final String CX = "ctxt:";
}
